#ifndef h_ocrinterfaces__
#define h_ocrinterfaces__

// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: iupr common header files
// File: ocrinterfaces.h
// Purpose: interfaces to OCR system components
// Responsible: tmb
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#ifdef SWIG
%module ocropus 
#endif

#include <stdlib.h>
#include "narray.h"
#include "misc.h"
#include "coords.h"
#include "idmap.h"

namespace iupr {
    struct IComponent {
	virtual ~IComponent() {}
	virtual const char *description() = 0;
	virtual void init(const char **argv=0) = 0;
    };

    struct ICleanupGray : IComponent {
	virtual void cleanup(bytearray &out,bytearray &in) = 0;
    };

    struct ICleanupBinary : IComponent {
	virtual void cleanup(bytearray &out,bytearray &in) = 0;
    };

    struct ISegmentPage : IComponent {
	virtual void segment(intarray &out,bytearray &in) = 0;
    };

    struct ISegmentLine : IComponent {
	virtual void charseg(intarray &out,bytearray &in) = 0;
    };

    struct IBinarize : IComponent {
	virtual void binarize(bytearray &out,floatarray &in) = 0;
    };

    struct IClassifyCharacter : IComponent {
	virtual void score(
	    intarray &classes,
	    floatarray &scores,
	    bytearray &in) = 0;
    };

    struct IClassifyLineCharacter : IComponent {
	virtual void score(
	    intarray &classes,
	    floatarray &scores,
	    bytearray &in,
	    int baseline,
	    int xheight,
	    int descender,
	    int ascender) = 0;
    };

    struct ILanguageModel : IComponent {
	/// start with an empty context
	virtual void start_context() = 0;
	/// start a new chunk for matching against the language model
	/// this may recognize things in the context of the previous text
	virtual void start_chunk(int nstates) = 0;
	/// add a transition 
	virtual void add_transition(int from,int to,int output,float cost,int id) = 0;
	/// get a special code for font changes, special characters, etc.
	virtual int special(const char *s) = 0;
	/// compute the nbest-matches
	virtual void compute(int n) = 0;
	/// return the actual number of matches
	virtual int nresults() = 0;
	// retrieve the nbest matches
	virtual void nbest(intarray &result,floatarray &costs,intarray &ids,intarray &states,int index) = 0;
        // destructor
        virtual ~ILanguageModel() {}
    };

    // FIXME: pretty up the comments and spacing and all that

    struct ISegmentedLineOCR : IComponent {
	// you must call this method before doing any kind of recognizing; if this is used as part of an
	// IRetrainableOCR object, this may take a long time to complete, since the actual training may be
	// carried out at this point
	virtual void start_recognizing() = 0;
	// the input is just given as the "segmentation" argument; on output, the OCR engine should put
	// its result into the language model "result" and map the components to their corresponding segments
	// through "idmap".
	//
	// The OCR builds "hypothesis graph" from the given segments.
	// The graph here is represented as a set of numbered arcs (transitions)
	// and is built inside the language model.
	// Every transition of the graph uses several segments,
	// and is about recognizing the shape of the union of those segments.
	// The relationship between transition ids and segments is returned
	// in the `components' idmap. 
	virtual void recognize_binary(ILanguageModel &result,idmap &components,intarray &segmentation) = 0;
	// FIXME same as above, but using segmentation as a mask for a grayscale image
	virtual void recognize_gray(ILanguageModel &result,idmap &components,intarray &segmentation,bytearray &image) = 0;
	// FIXME maybe add a "get character training images" method here
    };

    // FIXME: pretty up the comments and spacing and all that

    struct ISegmentedRetrainableOCR : ISegmentedLineOCR {
	virtual void cost_sample(floatarray &costs) = 0;
	// call this before starting training; you can restart training after recognizing if you
	// like; you may only call the train_* methods after calling start_training but before 
	// start_recognizing
	virtual void start_training() = 0;
	// chars is a list of characters, costs is the cost associated with each character,
	// segmentation is the color-coded segmentation; the colors represent the characters in the 
	// input string, with color=1 corresponding to character=0
	virtual void train_binary(ucarray &chars,floatarray &costs,intarray &charsegmentation) = 0;
	// same as train_binary, but the actual image data is in the image array, while the segmentation
	// just segments the image data into regions (that is, the segmentation itself need not look very
	// much like the characters and is, in general, at least several pixels wider than the binarized
	// grayscale characters)
	virtual void train_gray(ucarray &chars,floatarray &costs,intarray &charsegmentation,bytearray &image) = 0;
    };

    // There are many old-style OCR systems that are incapable of returning lattices or language models
    // and that are incapable of taking advantage of segmentation input.  These are interfaces to those
    // kinds of OCR systems.  They can be adapted to the language-based model interfaces using implementations
    // of the IAdapt... interfaces.
    // Both Tesseract and HMM-based recognizers should fit easily into these models (although the HMM-based
    // system is capable of generating language model output and segmentation information, so it might also
    // fit with the above interfaces)

    struct ISimpleLineOCR : IComponent {
	virtual void start_recognizing() = 0;
	// recognize the text line given by the binary image bimage
	virtual void recognize_binary(ucarray &result,floatarray &costs,narray<rectangle> &bboxes,bytearray &bimage) = 0;
	// recognize the text line given by the gray scale image image
	virtual void recognize_gray(ucarray &result,floatarray &costs,narray<rectangle> &bboxes,bytearray &image) = 0;
	// FIXME maybe add a "get character training images" method here
    };

    struct ISimpleRetrainableOCR : ISimpleLineOCR {
	virtual void start_training() = 0;
	// isolated character training: csegmentation contains masks for each character, starting with
	// color=1 for chars[0]; in the grayscale image, the segmentation is used as a mask for the image
	// note: if there are "bad" or "uncertain" characters in the training image, just leave them out
	// of the segmentation altogether
	virtual bool supports_char_training() = 0;
	virtual void train_binary_chars(ucarray &chars,intarray &csegmentation) = 0;
	virtual void train_gray_chars(ucarray &chars,intarray &csegmentation,bytearray &image) = 0;
	// unsegmented line training: given the input image and its transcription, the OCR system tries
	// to improve itself
	// note: if there are a significant number of "bad" or "uncertain" images in the training image,
	// perhaps it shouldn't be used for training at all, or the OCR system has to make that determination
	// itself (this interface doesn't care--that's something that needs to be worked out later)
	virtual bool supports_line_training() = 0;
	virtual void train_binary(ucarray &chars,bytearray &bimage) = 0;
	virtual void train_gray(ucarray &chars,bytearray &image) = 0;
    };

#if 0
    // these don't need to be defined as interfaces, but at least one implementation should
    // be provided somewhere (maybe multiple ones--there are different kinds of strategies possible)
    struct AdaptSimpleOCR : ILineOCR {
	// this is just here as an example
    };

    struct AdaptSimpleRetrainableOCR : ISimpleRetrainableOCR {
	// this is just here as an example
    };
#endif

}

#endif
