// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-adaptive
// File: langmod-aspell.cc
// Purpose: trivial WFST implementation that calls aspell
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#include <ctype.h>
#include "iupr.h"
#include "io_png.h"
#include "io_pbm.h"
#include "img.h"
#include "hash.h"
#include "ocrinterfaces.h"
#include "quicksort.h"
#include "process.h"
extern "C" {
    #include <aspell.h>
}

using namespace iupr;

/// This language model is a wrapper for another language model.
/// In the building phase, it's completely transparent;
/// when it produces the resulting string and costs,
/// it checks all words with ISpell and changes the cost of unrecognized words to 100.
/// Also the unrecognized words are replaced with some guess that is in the dictionary
/// (whether the guess is reasonable or not).

struct ASpellLanguageModel : ILanguageModel {
    autodel<ILanguageModel> langmod;
    AspellSpeller *spell_checker;

    void init_aspell() {
        AspellConfig *spell_config = new_aspell_config();
        aspell_config_replace(spell_config, "lang", "en_US");
        AspellCanHaveError *possible_err = new_aspell_speller(spell_config);
        delete_aspell_config(spell_config);
        spell_checker = NULL;
        if (aspell_error(possible_err))
            throw aspell_error_message(possible_err);
        else
            spell_checker = to_aspell_speller(possible_err);
    }

    void delete_aspell() {
         delete_aspell_speller(spell_checker);
    }

    virtual const char *description() {
        return "SpellLanguageModel";
    }

    virtual void init(const char **) {
    }


    void start_context() {
        langmod->start_context();
    }

    void start_chunk(int n) {
        langmod->start_chunk(n);
    }
    void add_transition(int from,int to,int output,float cost,int id) {
        langmod->add_transition(from, to, output, cost, id);
    }
    int special(const char *s) {
        return langmod->special(s);
    }
    void compute(int n) {
	langmod->compute(n);
    }

    inline bool word_continuation(int c) {
        return isalpha(c) || c == '\'';
    }


    // This method is way too long; sorry.
    void spellcheck(intarray &str,
                    floatarray &costs,
                    intarray &ids,
                    intarray &states,
                    float coef_good) {
        intarray new_classes;
        floatarray new_costs;
        intarray new_ids;
        intarray new_states;
        int word_start = 0;

        while(1) {
            // Move `word_start' to the start of the next word.
            while(word_start < str.length() && !isalpha(str[word_start])) {
                new_classes.push(str[word_start]);
                new_costs.push(costs[word_start]);
                new_ids.push(ids[word_start]);
                new_states.push(states[word_start]);
                word_start++;
            }

            if(word_start == str.length())
                break;

            int word_end = word_start;
            while(word_end < str.length() && word_continuation(str[word_end])) {
                word_end++;
            }

            // not checking one-letter words
            if(word_end == word_start + 1) {
                new_classes.push(str[word_start]);
                new_costs.push(costs[word_start]);
                new_ids.push(ids[word_start]);
                new_states.push(states[word_start]);
                word_start++;
                continue;
            }

            // Now put the word into Aspell.
            char buf[word_end - word_start + 1];
            for(int i = word_start; i < word_end; i++) 
                buf[i - word_start] = str[i];
            buf[word_end - word_start] = 0;
            int correct = aspell_speller_check(spell_checker, buf, word_end - word_start);

            if(correct) {
                // a good word - just copy to output arrays without problems
                for(int i = word_start; i < word_end; i++) {
                    new_classes.push(str[i]);
                    new_costs.push(costs[i] * coef_good);
                    new_ids.push(ids[i]);
                    new_states.push(states[i]);
                }
            } else {
                // a bad word: we want to get a possible correction
                const AspellWordList *suggestions = aspell_speller_suggest(spell_checker, buf, word_end - word_start);
                AspellStringEnumeration *aspell_elements = aspell_word_list_elements(suggestions);
                const char *corrected_word = aspell_string_enumeration_next(aspell_elements);

                // It was a requirement that a language model
                // should always output something in the language.
                // Even in the case when the word is completely unrecognizable!
                if(!corrected_word) {
                    corrected_word = "foo";
                }

                // We won't dive into questions,
                //     which letters were recognized bad and which were good.
                // Instead, we'll just give a huge cost to all the word,
                //     forbidding any adaptation on it.
                int len = strlen(corrected_word);
                for(int i = 0; i < len; i++) {
                    new_classes.push(corrected_word[i]);
                    new_costs.push(100);
                    new_ids.push(-1);
                    new_states.push(-1);
                }
                delete_aspell_string_enumeration(aspell_elements);
                
            }
            word_start = word_end;
        }

        move(str,   new_classes);
        move(costs,   new_costs);
        move(ids,       new_ids);
        move(states, new_states);
    }

    void nbest(intarray &results,
               floatarray &rcosts,
               intarray &rids,
               intarray &rstates,
               int index) {
        langmod->nbest(results, rcosts, rids, rstates, index);
        spellcheck(results, rcosts, rids, rstates, 0.1);
    }

    int nresults() {
        return langmod->nresults();
    }

    ASpellLanguageModel(ILanguageModel *lm) : langmod(lm) {
        init_aspell();
    }

    ~ASpellLanguageModel() {
        delete_aspell();
    }
};

namespace iupr {
    ILanguageModel *make_AspellLanguageModel(ILanguageModel *lm) {
	return new ASpellLanguageModel(lm);
    }
};
