// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-adaptive
// File: decorations.cc
// Purpose: rule-based detection of paragraphs and headers
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de


#include <stdlib.h>
#include <stdio.h>
#include "layout-rules.h"
#include "ocr-utils.h"

namespace iupr {

void detect_paragraphs(narray<bool> &result, narray<rectangle> &line_boxes) {
    makelike(result, line_boxes);
    if (!line_boxes.length())
        return;

    for(int i = 0; i < result.length() - 1; i++) {
        rectangle &rect = line_boxes[i];
        rectangle &next = line_boxes[i + 1];

        // A line is a paragraph starter if, compared to the next line:
        // it's above
        // its right edge is more or less the same
        // its left edge is shifted to the right
        result[i] = next.y0 < rect.y0
                 && abs(next.x1 - rect.x1) < 0.5 * rect.height()
                 && rect.x0 - next.x0 > 0.5 * rect.height();
    }
    // The last line is never a paragraph starter.
    if (result.length())
        result[result.length() - 1] = false;
}


void detect_sections(narray<bool> &result, narray<rectangle> &line_boxes) {
    makelike(result, line_boxes);
    if (!line_boxes.length())
        return;

    narray<bool> bigger_than_prev;
    narray<bool> bigger_than_next;
    makelike(bigger_than_prev, line_boxes);
    makelike(bigger_than_next, line_boxes);
    for(int i = 0; i < result.length(); i++) {
        int h = line_boxes[i].height();
        bigger_than_prev[i] = i == 0 || h > line_boxes[i-1].height() * 1.5;
        bigger_than_next[i] = i + 1 == result.length() 
                           || h > line_boxes[i+1].height() * 1.5;
    }

    

    /*intarray vskips;
    for (int i = 1; i < result.length(); i++) {
        if (line_boxes[i].y0 > line_boxes[i-1].y0) continue;
        vskips.push(line_boxes[i-1].y0 - line_boxes[i].y1);
    }
    int medskip = int(median(vskips));*/

    intarray skips;
    makelike(skips, line_boxes);
    skips[0] = 30000;
    for(int i = 1; i < result.length(); i++) {
        int vskip = line_boxes[i-1].y0 - line_boxes[i].y1;
        skips[i] = vskip;
    }

    const double afterskip_coef = .7;
    
    for(int i = 0; i < result.length(); i++) {
        result[i] = result.length() > 1 && (
            //   bigger_than_prev[i] && bigger_than_next[i]
            //|| i > 0 && bigger_than_prev[i-1] && bigger_than_next[i]
            //|| i + 1 < result.length() && bigger_than_prev[i] 
            //                           && bigger_than_next[i+1]
             
              // a line is a header if it has big skip before and medium afterwards
               i + 1 < result.length() 
                    && skips[i] > 1.5 * line_boxes[i].height() 
                    && skips[i + 1] > afterskip_coef * line_boxes[i].height()

              // same with 2-line headers: that's the rule for the first line
            || i + 2 < result.length() 
                    && skips[i] > 1.5 * line_boxes[i].height()
                    && skips[i + 2] > afterskip_coef * line_boxes[i+1].height()

              // and that's for the second line of 2-line headers
            || i > 0 
                    && i + 1 < result.length()
                    && result[i-1]
                    && skips[i + 1] > afterskip_coef * line_boxes[i].height()
        );
    }
}

};
