// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz
// or its licensors, as applicable.
//
// You may not use this file except under the terms of the accompanying license.
//
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Project: ocr-adaptive
// File: ocr-adapters.cc
// Purpose: implementation of adapters of ISimpleXXX to ISegmentedXXX
// Responsible: mezhirov
// Reviewer:
// Primary Repository:
// Web Sites: www.iupr.org, www.dfki.de

#include "ocr-utils.h"
#include "ocr-adapters.h"
#include "imglabels.h"
#include "io_png.h"

namespace iupr {

static int area(rectangle rect) {
    if (rect.width() <= 0 || rect.height() <= 0)
        return 0;
    return rect.width() * rect.height();
}

static bool fits_into(rectangle inner, rectangle outer) {
    if (outer.empty())
        return false;
    if (inner.empty())
        return false;
    rectangle common = inner.intersection(outer);
    return 2 * area(common) > area(inner);
}

// Given an output of SimpleOCR (string, costs, bboxes) and the segmentation,
// produce a SegmentedOCR results (result, components).
//
// This uses heuristical match between bounding boxes returned by OCR and
// the segmentation bounding boxes. That might fail, especially in the case
// of undersegmentation.
static void convert_results(ILanguageModel &result,
                            idmap &components,
                            ucarray &string,
                            floatarray &costs,
                            narray<rectangle> &bboxes,
                            intarray &segmentation) {
    check_line_segmentation(segmentation);
    white_to_black(segmentation);
    set_line_number(segmentation, 0);
    components.clear();
    int n = string.length();
    ASSERT(costs.length() == n);
    ASSERT(bboxes.length() == n);

    narray<rectangle> seg_boxes;
    bounding_boxes(seg_boxes, segmentation);

    result.start_chunk(n + 1);
    for(int i=0; i<n; i++) {
        result.add_transition(i, i + 1, string[i], costs[i], i);
        
        // fill components
        for(int j=1; j<seg_boxes.length(); j++) {
            if(fits_into(seg_boxes[j], bboxes[i])) {
                components.bind(i, j);
            }
        }
    }
}

/// This class is abstract because of get_simple_line_ocr(); this is done because
/// we want multiple inheritance: AdaptRetrainable would inherit both from AdaptSimple
/// and ISegmentedRetrainableOCR. But virtual inheritance requires changes into the interfaces,
/// and this is a little workaround.
/// Oh, multiple inheritance is also forbidden for non-interfaces by CCs.
class AbstractSimpleToSegmentedAdapter : public ISegmentedLineOCR {
public:
    virtual ISimpleLineOCR &get_simple_line_ocr() = 0;

    virtual const char *description() {
        return "Adapter of 'Simple' OCR to Segmented";
    }

    virtual void init(const char **argv=0) {
    }

    virtual void start_recognizing() {
        get_simple_line_ocr().start_recognizing();
    }

    

    virtual void recognize_binary(ILanguageModel &result,
                                  idmap &components,
                                  intarray &orig_segmentation) {
        intarray segmentation;
        copy(segmentation, orig_segmentation);
        check_line_segmentation(segmentation);
        ucarray string;
        floatarray costs;
        narray<rectangle> bboxes;
        bytearray image;
        forget_segmentation(image, segmentation);

	get_simple_line_ocr().recognize_binary(string, costs, bboxes, image);
        convert_results(result, components, string,
                        costs, bboxes, segmentation);
    }
    
    virtual void recognize_gray(ILanguageModel &result,
                                idmap &components,
                                intarray &orig_segmentation,
                                bytearray &image) {
        intarray segmentation;
        copy(segmentation, orig_segmentation);
        check_line_segmentation(segmentation);
        ucarray string;
        floatarray costs;
        narray<rectangle> bboxes;
	get_simple_line_ocr().recognize_gray(string, costs, bboxes, image);
        convert_results(result, components, string,
                        costs, bboxes, segmentation);
    }

};


// this is what make_adapter(ISimpleLineOCR &) creates; see the header.
class AdaptSimpleToSegmentedLineOCR : public AbstractSimpleToSegmentedAdapter {
public:
    autodel<ISimpleLineOCR> simple_line_ocr;

    virtual ISimpleLineOCR &get_simple_line_ocr() {
        return *simple_line_ocr;
    }

    AdaptSimpleToSegmentedLineOCR(ISimpleLineOCR *s) : simple_line_ocr(s) {
    }
};

/// Recolor the given segmentation leaving only segments
/// whose 0-based indices are listed in the given list.
/// The given list becomes the mapping from new 0-based indices to old ones.
static void recolor_segmentation(intarray &result,
                                 intarray &seg,
                                 intarray &map) {
    int n = max(seg);
    ASSERT(max(map) < n);
    intarray back_map;
    back_map.resize(n + 1);
    fill(back_map, 0);
    for(int i = 0; i < map.length(); i++) {
        back_map[map[i] + 1] = i + 1;
    }
    makelike(result, seg);
    for(int i = 0; i < seg.length1d(); i++)
        result.at1d(i) = back_map(seg.at1d(i));
}

/// Given a segmentation and a list of 0-based indices,
/// produce an image that includes only the specified segments.
static void image_from_segmentation(bytearray &result,
                                    intarray &seg,
                                    intarray &map) {
    int n = max(seg);
    bytearray back_map;
    back_map.resize(n + 1);
    fill(back_map, 255);
    for(int i = 0; i < map.length(); i++) {
        back_map[map[i] + 1] = 0;
    }
    makelike(result, seg);
    for(int i = 0; i < seg.length1d(); i++)
        result.at1d(i) = back_map(seg.at1d(i));
}


// this is what make_adapter(ISimpleRetrainableOCR &) creates; see the header.
class AdaptSimpleToSegmentedRetrainableOCR : public AbstractSimpleToSegmentedAdapter,
                                             public ISegmentedRetrainableOCR {
public:

    virtual const char *description() {return AbstractSimpleToSegmentedAdapter::description();}
    virtual void init(const char **argv=0) {AbstractSimpleToSegmentedAdapter::init();}
    virtual void start_recognizing() {AbstractSimpleToSegmentedAdapter::start_recognizing();}

    virtual void recognize_binary(ILanguageModel &result,
                                  idmap &components,
                                  intarray &segmentation) {
        AbstractSimpleToSegmentedAdapter::recognize_binary(result, components, segmentation);
    }
    
    virtual void recognize_gray(ILanguageModel &result,
                                idmap &components,
                                intarray &segmentation,
                                bytearray &image) {
        AbstractSimpleToSegmentedAdapter::recognize_gray(result, components, segmentation, image);
    }

    autodel<ISimpleRetrainableOCR> retrainable_line_ocr;
    float threshold; // characters with cost lower than this will be adapted to

    virtual void cost_sample(floatarray &a) {
        a.resize(1);
        a[0] = threshold;
    }
    virtual ISimpleLineOCR &get_simple_line_ocr() {
        return *retrainable_line_ocr;
    }
    AdaptSimpleToSegmentedRetrainableOCR(ISimpleRetrainableOCR *s,
                              float training_threshold)
        : retrainable_line_ocr(s), threshold(training_threshold) {
    }

    virtual void start_training() {
        retrainable_line_ocr->start_training();
    }
    
    /// Pick the characters that have cost less than the threshold and 
    /// collect their codes, costs and indices.
    void select_chars_for_adaption(ucarray &chars_to_adapt,
                                   floatarray &costs_to_adapt,
                                   intarray &map,
                                   ucarray &chars,
                                   floatarray &costs) {
        
        chars_to_adapt.clear();
        costs_to_adapt.clear();
        map.clear();

        int n = chars.length();
        ASSERT(n == costs.length());

        for(int i = 0; i < n; i++) {
            if (costs[i] > threshold) continue;
            chars_to_adapt.push(chars[i]);
            costs_to_adapt.push(costs[i]);
            map.push(i);
        }
    }

    virtual void train_binary(ucarray &chars,
                              floatarray &costs,
                              intarray &orig_charsegmentation) {
        intarray charsegmentation;
        copy(charsegmentation, orig_charsegmentation);
        check_line_segmentation(charsegmentation);
        white_to_black(charsegmentation);
        set_line_number(charsegmentation, 0);
        ASSERT(max(charsegmentation) <= chars.length());
        ASSERT(costs.length() == chars.length());
        ucarray new_chars;
        floatarray new_costs;
        intarray map;
        select_chars_for_adaption(new_chars, new_costs, map, chars, costs);
        if (!new_chars.length()) return;
        
        if (retrainable_line_ocr->supports_char_training()) {
            intarray new_segm;
            recolor_segmentation(new_segm, charsegmentation, map);
            black_to_white(new_segm);
            set_line_number(new_segm, 1);
            retrainable_line_ocr->train_binary_chars(new_chars, new_segm);
        } else if (retrainable_line_ocr->supports_line_training()) {
            bytearray new_image;
            image_from_segmentation(new_image, charsegmentation, map);
            retrainable_line_ocr->train_binary(chars, new_image);
        }
    }

    virtual void train_gray(ucarray &chars,
                            floatarray &costs,
                            intarray &orig_charsegmentation,
                            bytearray &image) {
        check_line_segmentation(orig_charsegmentation);
        intarray charsegmentation;
        copy(charsegmentation, orig_charsegmentation);
        ASSERT(max_cnum(charsegmentation) == chars.length());
        check_line_segmentation(charsegmentation);
        white_to_black(charsegmentation);
        set_line_number(charsegmentation, 0);
        ASSERT(costs.length() == chars.length());
        ucarray new_chars;
        floatarray new_costs;
        intarray map;
        select_chars_for_adaption(new_chars, new_costs, map, chars, costs);
        if (!new_chars.length()) return;
        
        if (retrainable_line_ocr->supports_char_training()) {
            intarray new_segm;
            recolor_segmentation(new_segm, charsegmentation, map);
            black_to_white(new_segm);
            set_line_number(new_segm, 1);
            retrainable_line_ocr->train_gray_chars(new_chars, new_segm, image);
        } else if (retrainable_line_ocr->supports_line_training()) {
            bytearray new_image;
            image_from_segmentation(new_image, charsegmentation, map);
            for(int i = 0; i < new_image.length1d(); i++) {
                if (!new_image.at1d(i))
                    new_image.at1d(i) = image.at1d(i);
            }

            retrainable_line_ocr->train_gray(chars, new_image);
        }
    }
};

ISegmentedLineOCR *make_adapter(ISimpleLineOCR *s) {
    return new AdaptSimpleToSegmentedLineOCR(s);
}

ISegmentedRetrainableOCR *make_adapter(ISimpleRetrainableOCR *s,
                                       float threshold) {
    return new AdaptSimpleToSegmentedRetrainableOCR(s, threshold);
}

};
