// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-adaptive
// File: ocr-recognize-page.cc
// Purpose: page reader: binarizes and makes two passes
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#include <time.h>
#include "iupr.h"
#include "checks.h"
#include "img.h"
#include "imgio.h"
#include "coords.h"
#include "imglabels.h"
#include "imgrescale.h"
#include "ocr-utils.h"
#include "imgio.h"
#include "triv-hocr.h"
#include "ocr-recognize-page.h"
#include "layout-rules.h"

namespace iupr {

param_bool zones("zones",false,"read UNLV zone files in <file name>.uzn");
param_string override_filename("override_filename",0,"print as file name regardless of the real file name");
param_bool allow_rescale("allow_rescale",true,"permit rescaling of line images to generate reasonably-sized images for the OCR engine");
param_bool orf("orf",false,"produce output in kooka/ocrad ORF format");
param_bool report("report", false, "report stages of OCR");
param_string debug_lines("debug_lines",0,
    "output each line, as found by the segmenter");
param_string debug_charseg_input("debug_charseg_input",0,
    "output each line before it's passed to the segmenter");
param_string debug_segm("debug_segm",0,"output segmentation file");
param_string debug_binarize("debug_binarize",0,
    "output the result of binarization");
param_string debug_retrain("debug_retrain",0,"output the data for retraining");
param_bool   debug_retrain_colormap("debug_retrain_colormap",0,
    "print table of recolorings for retraining");
param_int    adapt("adapt",1,"use adaption");
param_int    min_line_height("min_line_height",5,"minimum line height");
param_bool hocr("hocr",true,"produce output in hOCR format");
param_string charcut_file("charcut_file",0,"character cut file");
param_string charseg_file("charseg_file",0,"character segmentation file");
param_string transcript_file("transcript_file",0,"recognized characters file");
param_string train_costs_file("train_costs_file",0,"train costs file");

#include <sys/time.h>
#include <time.h>

static double now() {
    struct timeval time;
    gettimeofday(&time,0);
    return time.tv_sec + 0.000001 * time.tv_usec;
}

struct Timer {
    double total;
    double start;
    Timer() {
        total = 0.0;
        start = 0.0;
    }
    void reset() {
        total = 0.0;
        start = 0.0;
    }
    void operator+() {
        if(start!=0.0) throw "timer already running";
        start = now();
    }
    void operator-() {
        if(start==0.0) throw "timer not running";
        total += now() - start;
        start = 0.0;
    }
    double operator*() {
        if(start==0.0) return total;
        else return total + (now()-start);
    }
};

struct Timers {
    Timer binarizer,cleanup,page_segmenter,line_segmenter,ocr,langmod;
    void report() {
        fprintf(stderr,"time binarizer %g\n",*binarizer);
        fprintf(stderr,"time cleanup %g\n",*cleanup);
        fprintf(stderr,"time page_segmenter %g\n",*page_segmenter);
        fprintf(stderr,"time line_segmenter %g\n",*line_segmenter);
        fprintf(stderr,"time ocr %g\n",*ocr);
        fprintf(stderr,"time langmod %g\n",*langmod);
    }

    void reset() {
        binarizer.reset();
        cleanup.reset();
        page_segmenter.reset();
        line_segmenter.reset();
        ocr.reset();
        langmod.reset();
    }
};

static Timers ocr_timers;

void report_ocr_timings() {
    ocr_timers.report();
}

void reset_ocr_timings() {
    ocr_timers.reset();
}


static void prune(intarray &a, narray<bool> &mask) {
    intarray tmp;
    for(int i = 0; i < a.length(); i++) {
        if (mask[a[i]])
            tmp.push(a[i]);
    }
    move(a, tmp);
}

/// Prepare input for RetrainableOCR, using output of a SegmentedOCR.
/// @param recolor Output: recoloring map that maps segment
//                 indices into character indices (one character
//                 contains several segments).
/// @param to_train Output: classes of characters to train, in the order defined by recolor map.
/// @param costs_to_train Output: corresponding train costs.
/// @param boxes Output: bounding boxes of the characters.
static void gather_training_data_from_line(intarray &recolor,
                                           ucarray &to_train,
                                           floatarray &costs_to_train,
                                           ucarray &ocr_result,
                                           floatarray &costs,
                                           intarray &ids,
                                           idmap &im,
                                           narray<bool> &orig_used) {
    narray<bool> used;
    copy(used, orig_used);
    makelike(recolor, used);
    fill(recolor, 0);
    
    for(int i = 0; i < costs.length(); i++) {
        int id = ids[i];
        if (id >= 0 && ocr_result[i] != ' ') {
            intarray segs;
            im.segments_of_id(segs, id);

            prune(segs, used);
            
            if (!segs.length()) continue;
            to_train.push(ocr_result[i]);
            costs_to_train.push(costs[i]);
            int new_color = to_train.length();
            for(int k = 0; k < segs.length(); k++) {
                // This is a trick. The problem is that one segment might
                // belong to a different characters, but we only have one
                // recoloring slot. So we need to forbid more or less cleanly
                // recoloring a same segment twice, and therefore we're
                // going to maintain invariant:
                //
                // `used' holds information about segments used in characters
                // past the first i.
                used[segs[k]] = false;

                if(debug_retrain_colormap) {
                    printf("recoloring %d to %d\n",
                        segs[k], new_color);
                }
                if(segs[k] < recolor.length())
                    recolor[segs[k]] = new_color;
            }
        }
    }
}

static bool reject_line(bytearray &mask, bytearray &image) {
    ASSERT(samedims(mask, image));
    return image.dim(1) < min_line_height;
}

/// \brief Recognize one line given a pair of images (mask and gray),
///        and necessary components.
/// @param line Line index (for debugging output)
void recognize_line(ucarray &ocr_result,
                    ucarray &to_train,
                    floatarray &costs_to_train,
                    intarray &oversegmentation, 
                    intarray &true_segmentation,
                    bytearray &orig_line_mask,
                    bytearray &orig_gray_line,
                    ISegmentLine &line_segmenter,
                    ILanguageModel &langmod,
                    ISegmentedLineOCR &lineOCR,
                    int line) {
    if(reject_line(orig_line_mask, orig_gray_line)) {
        ocr_result.clear();
        to_train.clear();
        costs_to_train.clear();
        makelike(oversegmentation, orig_line_mask);
        makelike(true_segmentation, orig_line_mask);
        fill(oversegmentation, 0xFFFFFF);
        fill(true_segmentation, 0xFFFFFF);
        return;
    }

    bytearray line_mask;
    bytearray gray_line;
    copy(line_mask, orig_line_mask);
    copy(gray_line, orig_gray_line);
    ASSERT(samedims(line_mask, gray_line));
    if(allow_rescale) rescale_if_needed(line_mask, gray_line);
    ASSERT(samedims(line_mask, gray_line));

    intarray segline; // a segmentation of a possibly scaled line

    REQUIRE(line_mask.dim(0)>0 && line_mask.dim(1)>0);
    +ocr_timers.line_segmenter;
    line_segmenter.charseg(segline, check_narray_nomod<unsigned char>(line_mask));
    -ocr_timers.line_segmenter;
    check_line_segmentation(segline);
    REQUIRE(segline.dim(0)>0 && segline.dim(1)>0);
    REQUIRE(samedims(segline,line_mask));
    
    debug_write(debug_lines, line, segline);
    ASSERT(samedims(segline, line_mask));
    
    idmap im; 
    int max_segment = max(segline);
    im.set_max_segment(max_segment);
    ASSERT(samedims(segline, line_mask));
    +ocr_timers.ocr;
    lineOCR.recognize_gray(langmod,
                           im,
                           check_narray_nomod<int>(segline),
                           gray_line);
    -ocr_timers.ocr;
    //lineOCR.recognize_binary(langmod, im, segline);
    +ocr_timers.langmod;
    langmod.compute(1);
    -ocr_timers.langmod;

    ocr_result.clear();
    if(langmod.nresults()<1) {
        fprintf(stderr, "language model returns no result\n");
        return;
    }
    
    floatarray costs;
    intarray ids;
    intarray states;
    +ocr_timers.langmod;
    langmod.nbest(ocr_result, costs, ids, states, 0);
    -ocr_timers.langmod;
    ASSERT(samedims(segline, line_mask));

    //lineocr->recognize_binary(ocr,costs,lines[line]);
    
    // make oversegmentation
    makelike(oversegmentation, orig_line_mask);
    if(samedims(line_mask, orig_line_mask)) {
        move(oversegmentation, segline);
    } else {
        // rescale segmentation to its original size
        for(int x = 0; x < oversegmentation.dim(0); x++)
            for(int y = 0; y < oversegmentation.dim(1); y++) {
            oversegmentation(x, y) = segline(
                int(float(x) / oversegmentation.dim(0) * segline.dim(0)),
                int(float(y) / oversegmentation.dim(1) * segline.dim(1))
            );
        }
        normalize_segmentation(oversegmentation, orig_line_mask);
    }

    white_to_black(oversegmentation);
    set_line_number(oversegmentation, 0);
    // recolor to get true segmentation
    intarray recolor;
    narray<bool> used;
    used.resize(max_segment + 1);
    fill(used, false);
    for(int i = 0; i < oversegmentation.length1d(); i++)
        used[oversegmentation.at1d(i)] = true;
    gather_training_data_from_line(recolor, to_train, costs_to_train, ocr_result, costs, ids, im, used);

    makelike(true_segmentation, oversegmentation);
    for(int i = 0; i < oversegmentation.length1d(); i++)
        true_segmentation.at1d(i) = recolor[oversegmentation.at1d(i)];
    ASSERT(max(recolor) <= to_train.length());
    ASSERT(max(true_segmentation) <= to_train.length());

    // We can lose small segments due to rescaling down.
    // That means yet another recoloring.
    int color = 1;
    for(int i = 1; i < used.length(); i++) {
        if (used[i])
            recolor[i] = color++;
    }
    for(int i = 0; i < oversegmentation.length1d(); i++)
        oversegmentation.at1d(i) = recolor[oversegmentation.at1d(i)];
    black_to_white(segline);
    set_line_number(segline, 1);
    black_to_white(true_segmentation);
    set_line_number(true_segmentation, 1);
    black_to_white(oversegmentation);
    set_line_number(oversegmentation, 1);
    check_line_segmentation(segline);
    check_line_segmentation(true_segmentation);
    check_line_segmentation(oversegmentation);

    debug_write(debug_retrain, line, segline);
}


struct PageOCR : IPageOCR {
    autodel<IBinarize>                 binarizer;
    autodel<ICleanupGray>              cleanup_gray;
    autodel<ICleanupBinary>            cleanup_binary;
    autodel<ISegmentPage>              page_segmenter;
    autodel<ISegmentLine>              line_segmenter;
    autodel<ISegmentedLineOCR>         pass1;
    autodel<ILanguageModel>            langmod1;
    autodel<ISegmentedRetrainableOCR>  pass2;
    autodel<ILanguageModel>            langmod2;

    
    ~PageOCR() {
        if (pass1.ptr() == (ISegmentedLineOCR *)(pass2.ptr()))
            pass1.move();
        if (langmod1.ptr() == langmod2.ptr())
            langmod1.move();
    }

    void check_valid() {
         if (!binarizer) throw "binarizer not set";
         if (!page_segmenter) throw "page_segmenter not set";
         if (!line_segmenter) throw "line_segmenter not set";
         if (!pass1 && !pass2) throw "neither pass1 nor pass2 set";
         if (!langmod1 && !langmod2) throw "neither langmod1 nor langmod2 not set";
    }

    virtual void inject_binarizer(IBinarize *ptr) {
        if (!!binarizer) throw "binarizer already set";
        binarizer = ptr;
    }
    virtual void inject_cleanup_gray(ICleanupGray *ptr) {
        if (!!cleanup_gray) throw "cleanup-gray already set";
        cleanup_gray = ptr;
    }
    virtual void inject_cleanup_binary(ICleanupBinary *ptr) {
        if (!!cleanup_binary) throw "cleanup-gray already set";
        cleanup_binary = ptr;
    }
    virtual void inject_page_segmenter(ISegmentPage *ptr) {
        if (!!page_segmenter) throw "page_segmenter already set";
        page_segmenter = ptr;
    }
    virtual void inject_line_segmenter(ISegmentLine *ptr) {
        if (!!line_segmenter) throw "line_segmenter already set";
        line_segmenter = ptr;
    }
    virtual void inject_pass1(ISegmentedLineOCR *ptr) {
        if (!!pass1) throw "pass1 already set";
        pass1 = ptr;
    }
    virtual void inject_langmod1(ILanguageModel *ptr) {
        if (!!langmod1) throw "langmod1 already set";
        langmod1 = ptr;
    }
    virtual void inject_pass2(ISegmentedRetrainableOCR *ptr) {
        if (!!pass2) throw "pass2 already set";
        pass2 = ptr;
    }
    virtual void inject_langmod2(ILanguageModel *ptr) {
        if (!!langmod2) throw "langmod2 already set";
        langmod2 = ptr;
    }

    /// @param true_page_segmentation Character segmentation array,
    //                                will be modified only if non-empty.
    void pass(narray<ucarray> &ocr_result,
              narray<narray<rectangle> > &charboxes,
              narray<bytearray> &lines,
              narray<rectangle> &bboxes,
              ILanguageModel &langmod,
              ISegmentedLineOCR &lineOCR,
              bytearray &page,
              intarray &charcut,
              intarray &charseg,
              bool retrain_for_pass2) {
        stdio transcript;
        stdio train_costs;
        if(!retrain_for_pass2 && transcript_file)
            transcript = fopen(transcript_file, "w");
        if(!retrain_for_pass2 && train_costs_file)
            train_costs = fopen(train_costs_file, "w");
        langmod.start_context();
        objlist<bytearray> components;
        for(int line=0;line<lines.length();line++) {
            ASSERT(lines[line].dim(0) == bboxes[line].width());
            ASSERT(lines[line].dim(1) == bboxes[line].height());
            bytearray bin_line;
            copy(bin_line, lines[line]);
	    //invert(lines[line]);

            debug_write(debug_charseg_input, line, lines[line]);

            bytearray gray_line;
            extract_subimage(gray_line, page, bboxes[line].x0, bboxes[line].y0, bboxes[line].x1, bboxes[line].y1);

            ucarray to_train;
            floatarray costs_to_train;
            intarray oversegmentation;
            intarray true_segmentation;

            recognize_line(ocr_result[line], to_train, costs_to_train,
                           oversegmentation, true_segmentation,
                           bin_line, gray_line, *line_segmenter,
                           langmod, lineOCR, line);
            if(charcut.length1d()) {
                blit_segmentation_line(charcut, bboxes[line],
                                       oversegmentation, line);
            }
            if(charseg.length1d()) {
                blit_segmentation_line(charseg, bboxes[line],
                                       true_segmentation, line);
            }
            if(!!transcript) {
                for(int i = 0; i < to_train.length(); i++)
                    fprintf(transcript, "%lc", to_train[i]);
                fprintf(transcript, "%lc", '\n');
            }
            if(!!train_costs) {
                for(int i = 0; i < costs_to_train.length(); i++)
                    fprintf(train_costs, "%f\n", costs_to_train[i]);
                fprintf(train_costs, "\n");
            }
            
            if(retrain_for_pass2) {
                // FIXME: move training from here?
                //write_png_rgb(stdio("_debug_.png","w"),true_segmentation);
                check_line_segmentation(true_segmentation);
                pass2->train_gray(to_train, costs_to_train, true_segmentation, gray_line);
                //pass2->train_binary(to_train, costs_to_train, true_segmentation);
            }
        }    
    }

    /// Convert a binary image to a segmentation where white becomes background
    /// and black becomes noise (in accordance with the specs).
    /// (Hopefully non-noise pixels will be overwritten with other values).
    void prepare_segmentation(intarray &seg, bytearray &image) {
        makelike(seg, image);
        for(int i = 0; i < seg.length1d(); i++)
            seg.at1d(i) = image.at1d(i) ? 0xFFFFFF : 0xFFFF00;
    }

    void prepare_segmentations_if_needed(intarray &charcut,
                                         intarray &charseg,    
                                         bytearray &image) {
        if(charcut_file)
            prepare_segmentation(charcut, image);
        if(charseg_file)
            prepare_segmentation(charseg, image);
    }

    virtual void recognize_binarized(narray<ucarray> &ocr_result,
                                     narray<narray<rectangle> > &charboxes,
                                     narray<rectangle> &bboxes,
                                     bytearray &gray_image,
                                     bytearray &unclean_image) {
        ocr_result.clear();
        bboxes.clear();
        bytearray image;
        
        +ocr_timers.cleanup;
        if(!!cleanup_binary) {
            cleanup_binary->cleanup(image, unclean_image);
        } else {
            copy(image, unclean_image);
        }
        -ocr_timers.cleanup;

        intarray page;
        if (report) {     
             printf("segmenting page...");
             fflush(stdout);
        }
        +ocr_timers.page_segmenter;
        page_segmenter->segment(page, image);
        -ocr_timers.page_segmenter;
        if (report)
            printf("done\n");

        intarray rgb;
        narray<bytearray> lines;
        extract_lines(lines,bboxes,page);
        assert_all_items_nonempty(lines);

        ocr_result.resize(lines.length());
        charboxes.resize(lines.length());
        bboxes.resize(lines.length());
        intarray charseg;
        intarray charcut;
        if (!adapt)
            prepare_segmentations_if_needed(charcut, charseg, image);
            
        pass(ocr_result, charboxes, lines, bboxes,
             !langmod1 ? * langmod2 : *langmod1,
             !pass1 ? *pass2 : *pass1, gray_image,
             charcut, charseg, adapt);
        if(adapt) {
            prepare_segmentations_if_needed(charcut, charseg, image);
            ocr_result.renew(lines.length());
            pass(ocr_result, charboxes, lines, bboxes,
                 !langmod2 ? *langmod1 : *langmod2,
                 !pass2 ? *pass1 : *pass2, gray_image,
                 charcut, charseg, false);
        }
        if (charcut_file)
            write_png_rgb(stdio(charcut_file, "wb"), charcut);
        if (charseg_file)
            write_png_rgb(stdio(charseg_file, "wb"), charseg);
    }

    virtual void recognize_binary(narray<ucarray> &ocr_result,
                                  narray<narray<rectangle> > &charboxes,
                                  narray<rectangle> &bboxes,
                                  bytearray &image) {
        
        recognize_binarized(ocr_result, charboxes, bboxes, image, image);
    }

    virtual void recognize_gray(narray<ucarray> &lines, 
                                narray<narray<rectangle> > &charboxes,
                                narray<rectangle> &bboxes,
                                bytearray &image) {
        bytearray binarized;
        if(contains_only(image, byte(0), byte(255))) {
            copy(binarized, image);
        } else {
            floatarray temp;
            if(!!cleanup_gray) {
                bytearray cleaned;
                +ocr_timers.cleanup;
                cleanup_gray->cleanup(cleaned, image);
                -ocr_timers.cleanup;
                copy(temp, cleaned);
            } else {    
                copy(temp, image);
            }
            if (report) {     
                 printf("binarization...");
                 fflush(stdout);
            }
            +ocr_timers.binarizer;
            binarizer->binarize(binarized,temp);
            -ocr_timers.binarizer;
            if (report)
                printf("done\n");
        }
        recognize_binarized(lines, charboxes, bboxes, image, binarized);
    }

    void init(const char **) {}
    const char *description() {return "PageOCR";}
};

IPageOCR *make_PageOCR() {
    return new PageOCR();
}


static bool is_tick(int c) {
    return c == '`' || c == '\'';
}

static void output_text_line(FILE *file, ucarray &line) {
    for(int i = 0; i < line.length(); i++) {

        // '', ``, `', '` -> "
        if (is_tick(line[i]) && i + 1 < line.length() && is_tick(line[i+1])) {
            fprintf(file, "%lc", '"');
            i++;
        } else
            fprintf(file, "%lc", line[i]);
    }
    fputc('\n', file);
}


// Read an UNLV zone file (see UNLV ISRI OCRtk)
static void read_unlv_zones(narray<rectangle> &result, FILE *file, int height) {
    int x, y, w, h;
    char buf[1000];
    while (1) {
        if (!fgets(buf, sizeof(buf), file))
            break;
        if (sscanf(buf, "%d %d %d %d", &x, &y, &w, &h) != 4)
            break;
        int y1 = height - y - 1;
        int y0 = y1 - h;
        result.push(rectangle(x, y0, x + w, y1));
    }
}

static void output_orf_char(FILE *f, int c, rectangle &bbox, int image_height) {
    printf("%d %d %d %d;1,'%c'0\n", bbox.x0, image_height - bbox.y1 - 1, bbox.width(), bbox.height(), c);
}


static void output_result(narray<ucarray> &lines, 
                          narray< narray<rectangle> > &charboxes,
                          narray<rectangle> &bboxes,
                          bytearray &image) {
    if (hocr) {
        narray<bool> is_paragraph;
        narray<bool> is_header;
        detect_paragraphs(is_paragraph, bboxes);
        detect_sections(is_header, bboxes);

        for(int line = 0; line < lines.length(); line++) {
            if(is_paragraph[line])
                printf("<p class='ocr_par'>\n"); 
            if(!is_header[line]) {
                output_hOCR_line(stdout, lines[line], bboxes[line], image.dim(1));
            } else {
                if(line == 0 || !is_header[line-1])
                    printf("<h3>"); 
                output_hOCR_line(stdout, lines[line], bboxes[line], image.dim(1));
                if(line == lines.length() - 1 || !is_header[line + 1])
                    printf("</h3>\n"); 
            }
        }
    } else if (orf) {
        printf("total blocks 1\n");
        printf("block %d %d %d %d\n", 0, 0, image.dim(0), image.dim(1));
        printf("lines %d\n", lines.length());
        for(int line = 0; line < lines.length(); line++) {
            printf("line %d chars %d height %d\n", line, lines[line].length(), bboxes[line].height());
            for (int i = 0; i < lines[line].length(); i++) {
                output_orf_char(stdout, lines[line][i], charboxes[line][i], image.dim(1));
            }
        }        
    } else {
        for(int line = 0; line < lines.length(); line++)
            output_text_line(stdout, lines[line]);
    }
}


void recognize_multiple_pages_to_stdout(narray<strbuf> &files, IPageOCR &pageocr) {
    if (hocr) {
	printf("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">\n");
	printf("<html>\n");
        printf("<head>\n");
        printf("<meta name='ocr-system' content='OCRopus %s'>\n", get_version_string());  
        printf("<meta name='ocr-capabilities' content='ocr_line ocr_page'>\n");
        printf("<meta name='ocr-langs' content='en'>\n");
        printf("<meta name='ocr-scripts' content='Latn'>\n");
        printf("<meta name='ocr-microformats' content=''>\n");
	printf("<title>OCR Output</title>\n");
	printf("</head>\n");
        printf("<body>\n");
    }

    for(int pageno=0;pageno<files.length();pageno++) {
        const char *file_name = files[pageno];
        if (hocr)
            printf("<div class='ocr_page' title='image %s'>\n", override_filename ? override_filename : file_name);
        else if (orf)
            printf("source file %s\n", override_filename ? override_filename : file_name);
	bytearray image;
	//read_png(image,stdio(files[pageno],"r"),true);
	read_image_gray(image, files[pageno]);
        narray<ucarray> lines;
        narray<rectangle> bboxes;
        narray<narray<rectangle> > charboxes;

        clock_t begin = clock();

        if (zones) {
            strbuf zone_file_name;
            zone_file_name.ensure(strlen(file_name) + 10);
            char *last_dot = strrchr(file_name, '.');
            if (last_dot) *last_dot = '\0';
            strcpy(zone_file_name, file_name);
            if (last_dot) *last_dot = '.';
            strcat(zone_file_name, ".uzn");
            
            narray<rectangle> zone_boxes;
            read_unlv_zones(zone_boxes, stdio(zone_file_name,"r"), image.dim(1));

            // recognize every zone
            for (int i = 0; i < zone_boxes.length(); i++) {
                narray<ucarray> lines;
                narray<rectangle> bboxes;
                narray<narray<rectangle> > charboxes;
                bytearray zone_image;
                crop(zone_image, image, zone_boxes[i]);
                if (i==0)
                    write_png(stdio("zone0.png", "w"), zone_image);
                if (hocr) printf("<div class='ocr_column' title='%d %d %d %d'>\n",
                    zone_boxes[i].x0, image.dim(1) - zone_boxes[i].y1 - 1, 
                    zone_boxes[i].x1 - zone_boxes[i].x0, zone_boxes[i].y1 - zone_boxes[i].y0);
                pageocr.recognize_gray(lines, charboxes, bboxes, zone_image);
                output_result(lines, charboxes, bboxes, image /*yes, not zone image */);
                if (hocr) printf("</div>\n");
            }
        } else {
            pageocr.recognize_gray(lines, charboxes, bboxes, image);
            output_result(lines, charboxes, bboxes, image);
        }

        if (hocr) {
            printf("</div>\n");
            printf("<!-- Time: %.2f sec -->\n", double(clock() - begin) / CLOCKS_PER_SEC);
        }
    }

    if (hocr) {
        printf("</body>\n");
        printf("</html>\n");
    }
}


};
