// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-adaptive
// File: ocr-recognize-page.h
// Purpose: interface to the page reader
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de


#ifndef h_ocr_recognize_page_
#define h_ocr_recognize_page_

#include "ocrinterfaces.h"

namespace iupr {
    /// This is the high-level interface to the page reader.
    /// It wires together binarization, layout analysis,
    /// and 2 passes with retraining.
    /// It recognizes just text (header/paragraph detection is done afterwards).
    /// Methods in this class need only an image as input, other args are output.
    ///
    /// To use this object, you should inject the components with inject_XXX() functions.
    /// PageOCR has ownership of all those objects.
    /// If only one of pass1/pass2 or langmod1/langmod2 is injected, it will be used for both.
    /// The same result you may get by injecting the same object twice in these positions
    /// (deleting is correctly handled).
    ///
    /// Other objects should be injected for recognize_XXX() to work.
    /// Page recognition works like this:
    ///  - binarization (if needed)
    ///  - layout analysis (extracting lines)
    ///  - pass 1: recognizing all lines
    ///  - adaptation (probably interleaved with pass 2)
    ///    on characters recognized confidently enough
    ///    (confidently from the combined viewpoint of the OCR and language model)
    ///  - pass 2: recognizing all lines again with an adapted classifier.
    ///
    /// Don't assign stupid language models to pass 2 since they can spoil the text.

    struct IPageOCR : IComponent {
        virtual void recognize_binary(narray<ucarray> &lines,
                narray<narray<rectangle> > &charboxes,
                narray<rectangle> &lineboxes,
                bytearray &image) = 0;

        virtual void recognize_gray  (narray<ucarray> &lines,
                narray<narray<rectangle> > &charboxes,
                narray<rectangle> &lineboxes,
                bytearray &image) = 0;

        virtual void inject_binarizer(IBinarize *) = 0;
        virtual void inject_cleanup_gray(ICleanupGray *ptr) = 0;
        virtual void inject_cleanup_binary(ICleanupBinary *ptr) = 0;
        virtual void inject_page_segmenter(ISegmentPage *) = 0;
        virtual void inject_line_segmenter(ISegmentLine *) = 0;
        virtual void inject_pass1(ISegmentedLineOCR *) = 0;
        virtual void inject_langmod1(ILanguageModel *) = 0;
        virtual void inject_pass2(ISegmentedRetrainableOCR *) = 0;
        virtual void inject_langmod2(ILanguageModel *) = 0;
    };


    IPageOCR *make_PageOCR();
    void recognize_multiple_pages_to_stdout(narray<strbuf> &files, IPageOCR &pageocr);
    void recognize_line(ucarray &ocr_result,
            ucarray &to_train,
            floatarray &costs_to_train,
            intarray &oversegmentation, 
            intarray &true_segmentation,
            bytearray &line_mask,
            bytearray &gray_line,
            ISegmentLine &line_segmenter,
            ILanguageModel &langmod,
            ISegmentedLineOCR &lineOCR,
            int line);

    void report_ocr_timings();
    void reset_ocr_timings();
};

#endif
