// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz
// or its licensors, as applicable.
//
// You may not use this file except under the terms of the accompanying license.
//
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Project:
// File: process.cc
// Purpose: running processes
// Responsible: mezhirov
// Reviewer:
// Primary Repository:
// Web Sites: www.iupr.org, www.dfki.de

#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include "process.h"

namespace iupr {

static void checked_pipe(int filedes[2]) {
    if (pipe(filedes)) {
        perror("Cannot create pipe");
        exit(EXIT_FAILURE);
    }
}


struct Process : IProcess {
    FILE *m_into;
    FILE *m_from;
    pid_t m_pid;

    virtual FILE *into() {
        return m_into;
    }

    virtual FILE *from() {
        return m_from;
    }
    
    virtual pid_t pid() {
        return m_pid;
    }

    virtual void close() {
        if (!pid())
            return;
        if (m_from)
            fclose(m_from);
        if (m_into);
            fclose(m_into);
        int status;
        waitpid(pid(), &status, 0);
        m_pid = 0;
        m_from = NULL;
        m_into = NULL;
    }

    Process(const char *path, char *const argv[]) : 
        m_into(NULL), m_from(NULL), m_pid(0) {
        int into_prog[2];
        int from_prog[2];

        checked_pipe(into_prog);
        checked_pipe(from_prog);

        if ((m_pid = fork())) {
            ::close(into_prog[0]);
            ::close(from_prog[1]);
            m_from = fdopen(from_prog[0], "r");
            m_into = fdopen(into_prog[1], "w");
            if (!m_from || !m_into) {
                perror("Unable to use file descriptors");
                exit(1);
            }
            setbuf(m_from, NULL);
            setbuf(m_into, NULL);
        } else {
            ::close(into_prog[1]);
            ::close(from_prog[0]);
            dup2(into_prog[0], 0);
            dup2(from_prog[1], 1);
            execvp(path, argv);

            // if we're here, then execvp failed
            perror(path);
            exit(EXIT_FAILURE);
        }
    }

    virtual void sendEof() {
        fclose(m_into);
        m_into = NULL;
    }

    virtual ~Process() {
        close();
    }
};

IProcess *run_process(const char *path) {
    char const *argv[] = {path, NULL};
    return run_process(path, argv);
}

IProcess *run_process(const char *path, const char *argv[]) {
    return new Process(path, const_cast<char *const *>(argv) /* that's bad! */);
}

}
