// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz
// or its licensors, as applicable.
//
// You may not use this file except under the terms of the accompanying license.
//
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Project:
// File: process.h
// Purpose: running processes with pipes to both their stdin and stdout
// Responsible: mezhirov
// Reviewer:
// Primary Repository:
// Web Sites: www.iupr.org, www.dfki.de


#ifndef h_process_
#define h_process_

#include <stdio.h>
#include <sys/types.h>

namespace iupr {

/// A process communicating with the main program through a pair of pipes.
/// The current implementation requires POSIX functions such as
/// execve(), fork(), waitpid(), pipe(), dup2().
struct IProcess {
    /// Get the unbuffered (FILE *) object that sends data to stdin of the process.
    /// Don't fclose it!
    virtual FILE *into() = 0;

    /// Get the unbuffered (FILE *) object that receives data from stdout of the process.
    /// Don't fclose it!
    /// \warning Despite we don't buffer this stream, the process is likely to.
    /// \warning The stream is blocking, i.e. you might hang while reading from it.
    virtual FILE *from() = 0;

    /// Get the pid of the process.
    virtual pid_t pid() = 0;

    /// Send an EOF, i.e. close the into() file.
    virtual void sendEof() = 0;

    /// Wait for the process to terminate.
    virtual void close() = 0;

    /// Wait for the process to terminate and destroy the object.
    virtual ~IProcess() {}
};

IProcess *run_process(const char *path);
IProcess *run_process(const char *path, const char *argv[]);

};

#endif
