// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project:
// File: 
// Purpose: 
// Responsible: 
// Reviewer: 
// Primary Repository: 
// Web Sites: 

#include "iupr.h"
#include "imgio.h"
#include "ocr-utils.h"
#include "tesseract.h"
#include "ocr-adapters.h"

using namespace iupr;


/// This language model accepts only transitions from i to i+1,
/// thus the hypotheses graph contains a single path.
/// That's useless for anything except testing.
struct SinglePathLanguageModel : ILanguageModel {
    ucarray text;
    floatarray costs;
    
    virtual const char *description() {
        return "SinglePathLanguageModel";
    }
    
    virtual void init(const char **) {
    }
    
    virtual void start_context() {
    }

    virtual void start_chunk(int n) {
        ASSERT(n > 0);
        text .resize(n - 1);
        costs.resize(n - 1);
        fill(costs, 1e10);
    }

    virtual void add_transition(int from, int to, int output, float cost, int id) {
        ASSERT(to == from + 1);
        ASSERT(id == from);
        if (cost < costs[from]) {
            text[from] = output;
            costs[from] = cost;
        }
    }

    virtual int special(const char *s) {
        return 0;
    }

    virtual void compute(int n) {
    }
    
    virtual int nresults() {
        return 1;
    }
    
    virtual void nbest(intarray &result, floatarray &costs, intarray &ids, intarray &states, int index) {
        ASSERT(index == 0);
        copy(result, text);
        copy(costs, this->costs);
        makelike(ids, result);
        for(int i = 0; i < ids.length(); i++)
            ids[i] = i;
        states.resize(result.length() + 1);
        for(int i = 0; i < states.length(); i++)
            states[i] = i;
    }
    
};


void test_ocr_result(ILanguageModel &langmod, const char *gt) {
    intarray ocr_result;
    floatarray costs;
    intarray ids;
    intarray states;
    langmod.nbest(ocr_result, costs, ids, states, 0);
    int gt_len = strlen(gt);
    for(int i = 0; i < ocr_result.length(); i++) {
        if(i >= gt_len || ocr_result[i] != gt[i]) {
            fprintf(stderr, "mistaken char #%d for %lc in %s", i, ocr_result[i], gt);
        }
    }
}

static void test_line_ocr(ISegmentedLineOCR &ocr) {
    stdio filelist("test-ocr-adapters.list", "r");
    ocr.start_recognizing();
    char in_path[1000];
    while(fscanf(filelist, "%999s", in_path) == 1) {
        char seg_path[1000];
        fscanf(filelist, "%999s", seg_path);
        char gt_path[1000];
        fscanf(filelist, "%999s", gt_path);
        char gt[1000];
        stdio gt_file(gt_path, "r");
        fgets(gt, sizeof(gt), gt_file);
        bytearray image;
        SinglePathLanguageModel langmod;
        idmap im;
        narray<rectangle> bboxes;
        read_image_gray(image, in_path);
        intarray seg;
        read_png_rgb(seg, stdio(seg_path, "rb"));
        ocr.recognize_gray(langmod, im, check_narray_nomod<int>(seg),
                                        check_narray_nomod<byte>(image));
        test_ocr_result(langmod, gt);
        binarize_by_threshold(image);
        ocr.recognize_binary(langmod, im, check_narray_nomod<int>(seg));
        test_ocr_result(langmod, gt);
    }
}

int main(int argc, char **argv) {
    autodel<ISegmentedRetrainableOCR> ocr(make_adapter(make_tesseract(argv[0]), 30));
    ocr->description();
    ocr->init(NULL);
    intarray alice;
    read_png_rgb(alice, stdio("alice.png", "rb"));
    ASSERT(alice.rank() == 2);
    bytearray alice_bin;
    forget_segmentation(alice_bin, alice);
    ucarray alice_text(5);
    alice_text[0] = 'A';
    alice_text[1] = 'l';
    alice_text[2] = 'i';
    alice_text[3] = 'c';
    alice_text[4] = 'e';
    floatarray costs;
    costs.resize(5);
    fill(costs, 1);
    check_line_segmentation(alice);
    try{ocr->train_binary(alice_text, costs, check_narray_nomod<int>(alice)); CHECK(0);} catch (...) {}
    try{ocr->train_gray  (alice_text, costs, check_narray_nomod<int>(alice), alice_bin); CHECK(0);} catch (...) {}
    test_line_ocr(*ocr);
    ocr->start_training();
    ocr->train_binary(alice_text, costs, check_narray_nomod<int>(alice));
    ocr->train_gray  (alice_text, costs, check_narray_nomod<int>(alice), alice_bin);
    ocr->start_recognizing();
    test_line_ocr(*ocr);
}

