// -*- C++ -*-

// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-layout-rast - Extract textlines from a document image using RAST
// File: ocr-char-stats.cc
// Purpose: calculate character statistics from bounding boxes
// Responsible: Faisal Shafait (faisal.shafait@dfki.de)
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#include "iupr.h"
#include "io_png.h"
#include "io_pbm.h"
#include "img.h"
#include "ocr-layout-rast.h"

using namespace iupr;

namespace iupr {
    extern param_int debug;

    CharStats::CharStats() {
        img_height    = -1;
        img_width     = -1;
        xheight       = -1;
        char_spacing  = -1;
        word_spacing  = -1;
        line_spacing  = -1;
        char_boxes.clear();
        dot_boxes.clear();
    }

    CharStats::CharStats(CharStats &c) {
        img_height    = c.img_height;
        img_width     = c.img_width;
        xheight       = c.xheight;
        char_spacing  = c.char_spacing;
        word_spacing  = c.word_spacing;
        line_spacing  = c.line_spacing;
        copy(char_boxes,c.char_boxes);
        copy(dot_boxes,c.dot_boxes);
    }

    CharStats::~CharStats() {
        char_boxes.dealloc();
        dot_boxes.dealloc();
    }

    void CharStats::get_char_boxes(narray<rectangle> &concomps){

        if(!concomps.length()){
            fprintf(stderr,"Empty connected components array given to compute character statistics. Aborting ...\n");
            exit(01);
        }
        img_width  = concomps[0].x1;
        img_height = concomps[0].y1;
        int histsize  = 200;
        int mindim    = 5;
	floatarray xhist(histsize),yhist(histsize);
        fill(xhist,0.0);
        fill(yhist,0.0);
	for(int i=1;i<concomps.length();i++) {
            int w = concomps[i].width();
            int h = concomps[i].height();
            if(w<=mindim || h<=mindim || w>=histsize-1 || h>=histsize-1 )
                continue;
            xhist[w]++;
            yhist[h]++;
	}
        //plot_hist(fopen("xhist-original.png","w"),xhist);
        //plot_hist(fopen("yhist-original.png","w"),yhist);
	//gauss1d(xhist,1.0);
	//gauss1d(yhist,1.0);
        //plot_hist(fopen("xhist.png","w"),xhist);
        //plot_hist(fopen("yhist.png","w"),yhist);
        int xmode = argmax(xhist);
        int ymode = argmax(yhist);
        if(debug>=2){
            printf("x-mode = %d, y-mode =%d, x-peak = %.2f, y-peak=%.2f\n",
                   xmode,ymode,xhist[xmode],yhist[ymode]);
        }

        int minarea    = ymode*2;
        int mindotarea = ymode/3;
        int maxaspect  = 10;
        int maxheight  = ymode*10;
        int maxwidth   = xmode*10;
        int minheight  = ymode>>1;
        int minwidth   = xmode>>1;
 	for(int i=1;i<concomps.length();i++) {
 	    int bw = concomps[i].width();
 	    int bh = concomps[i].height();
 	    if(bw<=0 || bh<=0) continue;
            if(bw*bh<minarea){
                if(bw*bh>mindotarea)
                    dot_boxes.push(concomps[i]);
                continue;
            }
 	    if(bw<minwidth && bh<minheight){
                dot_boxes.push(concomps[i]);
                continue;
            }
 	    if(bw/bh>maxaspect || bh/bw>maxaspect) continue;
 	    if(bw>maxwidth) continue;
 	    if(bh>maxheight) continue;
 	    char_boxes.push(Rect(concomps[i]));
 	}

    }
    int compare_rects_by_x0(const void *va,const void *vb) {
        const Rect *a = (const Rect*)va;
        const Rect *b = (const Rect*)vb;
        if(a->x0<b->x0) return -1;
        if(a->x0>b->x0) return 1;
        return 0;
    }
    
    void CharStats::sort_boxes_by_x0(narray<Rect> &boxes) {
        qsort(&boxes[0],boxes.length(),sizeof boxes[0],compare_rects_by_x0);
    }
    
    int compare_rects_by_y0(const void *va,const void *vb) {
        const Rect *a = (const Rect*)va;
        const Rect *b = (const Rect*)vb;
        if(a->y0<b->y0) return -1;
        if(a->y0>b->y0) return 1;
        return 0;
    }
    
    void CharStats::sort_boxes_by_y0(narray<Rect> &boxes) {
        qsort(&boxes[0],boxes.length(),sizeof boxes[0],compare_rects_by_y0);
    }
    
    void CharStats::calc_hist(floatarray &distances, floatarray &hist){
        
        float minval=10000;
        float maxval=-10000;
        int   len = distances.length();
        for(int i=0; i<len; i++){
            maxval = max(maxval, distances[i]);
            minval = min(minval, distances[i]);
        }
        
        //cerr<< "MaxVal = "<<maxval;
        
        hist.clear();
        hist.resize((int)maxval+1); //Not interested in -ive values
        //cerr <<"\tHistlen = "<<hist.length()<<endl;
        for(int i=0; i<=maxval; i++)
            hist[i] = 0;
        for(int i=0; i<len; i++){
            hist[(int)distances[i]] ++;
        }
        //cerr <<"\tHistlen = "<<hist.length()<<endl;
    }
    
    inline bool CharStats::x_overlap (Rect a, Rect b){
        return ( (a.x1 >= b.x0) && (b.x1 >= a.x0) );
    }
    
    inline bool CharStats::y_overlap (Rect a, Rect b){
        return ( (a.y1 >= b.y0) && (b.y1 >= a.y0) );
    }
    
    /* estimate the height of a line of text using bouding boxes of connected components */
    int CharStats::estimate_xheight(narray<Rect> &bboxes) {
        
        int i;
        int best_i;
        float best_v;
        floatarray heights;
        floatarray histogram;
        int h=200, min_height=2;
        float smooth=4.0;
        int len = bboxes.length();
        for(int i=0; i< len; i++){
            if(bboxes[i].height()<h)
                heights.push(bboxes[i].height());
        }
        calc_hist(heights,histogram);
        gauss1d(histogram,smooth);
        
        best_i = -1;
        best_v = -1.0;
        for(i=min_height;i<histogram.length();i++) {
            if(histogram[i]<best_v) continue;
            best_i = i;
            best_v = histogram[i];
        }
        
        return best_i;
    }
    
    void CharStats::calc_char_stats(){
        calc_char_stats(char_boxes);
    }
    
    void CharStats::calc_char_stats(narray<Rect> &cboxes){
        
        int num_chars      = cboxes.length();
        if(!cboxes.length()){
            if(debug)
                fprintf(stderr,"No character boxes found! ...\n");
            return ;
        }
        
        xheight = estimate_xheight(cboxes);
        sort_boxes_by_y0(cboxes);
        
        narray<Rect> line;
        floatarray   dist,vdist,hdist;
        floatarray   hist,vhist,hhist;
        float        maxdist = 200;
        float        temp=0,minhdist,minvdist;
        int          start;
        float        smooth = 2.0; 
        
        //Start with i=1 to skip first bounding box which is page size.
        start = ( (cboxes[0].x0==0) && (cboxes[0].y0==0) ) ? 1 : 0;
        for(int i=start; i< num_chars; i++){
            Rect r = cboxes[i];
            minhdist = 10000;
            minvdist = 10000;
            for(int j=start; j< num_chars; j++){
                Rect cand = cboxes[j];
                if(y_overlap(r,cand) && (cand.x0 > r.x1)){
                    temp = cand.x0 - r.x1;
                    if(temp < minhdist)
                        minhdist = temp;
                }
                if(x_overlap(r,cand) && (cand.y0 > r.y1)){
                    temp = cand.y0 - r.y1;
                    if(temp < minvdist)
                        minvdist = temp;
                }
            }
            if(minhdist && (minhdist<maxdist))
                hdist.push(minhdist);
            if(minvdist && (minvdist<maxdist))
                vdist.push(minvdist);
        }
        calc_hist(hdist, hhist);
        calc_hist(vdist, vhist);
        gauss1d(hhist,smooth);
        gauss1d(vhist,smooth);
        //plot_hist(fopen("char_hhist.png","w"),hhist);
        //plot_hist(fopen("char_vhist.png","w"),vhist);
        intarray modes;
        peaks(modes, hhist, 1, int(maxdist));
        if(modes.length()<2){
            if(debug)
                fprintf(stderr,"Warning: No peaks found in character spacing histogram!\n");
            char_spacing = xheight / 4;
            word_spacing = xheight / 2;
        }
        else{
            char_spacing = modes[0];
            word_spacing = modes[1];
        }
        float mv = -1;
        for(int i=0, l=vhist.length(); i<l; i++)
            if(vhist[i] > mv){
                mv = vhist[i];
                temp = i;
            }
        if(mv > 0)
            line_spacing = (int) temp;
    }
    
    void CharStats::print() {
        printf("Image Size   = %d x %d\n",img_width,img_height);
        printf("xheight      = %d\n",xheight);
        printf("Char spacing = %d\n",char_spacing);
        printf("Word spacing = %d\n",word_spacing);
        printf("Line spacing = %d\n",line_spacing);
        printf("Number of Characters = %d\n",char_boxes.length());
        printf("Number of dots = %d\n",dot_boxes.length());
    }
    CharStats *make_CharStats() {
        return new CharStats();
    }
    CharStats *make_CharStats(CharStats &c) {
        return new CharStats(c);
    }
}
