// -*- C++ -*-

// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-layout-rast - Extract textlines from a document image using RAST
// File: ocrctextlinerast.h
// Purpose: Header file declaring data structures for constrained textline
//          extraction using RAST
// Responsible: Faisal Shafait (faisal.shafait@dfki.de)
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#ifndef h_ocrctextlinerast__
#define h_ocrctextlinerast__

#include "ocrinterfaces.h"
#include "rect.h"
#include "struct.h"
#include "heap.h"
#include "iarith.h"
#include "iupr.h"

#include "ocr-char-stats.h"

namespace iupr {

    /////////////////////////////////////////////////////////////////////
    ///
    /// \struct TextLine
    /// Purpose: Textline bounding box and it attributes
    ///
    //////////////////////////////////////////////////////////////////////
    
    struct TextLine {
        float c,m,d; // c is y-intercept, m is slope, d is the line of descenders
        int   xheight;
        Rect  bbox;
        void print(){
            printf("%d %d %d %d ",bbox.x0,bbox.y0,bbox.x1,bbox.y1);
            printf("%.3f %f %.2f %d\n",c,m,d,xheight);
        }
    };

    /////////////////////////////////////////////////////////////////////
    ///
    /// \struct CTextlineRAST
    /// Purpose: Constrained Textline finding using RAST
    ///
    //////////////////////////////////////////////////////////////////////
    
    const int nparams = 3;
    class CTextlineRAST {
    private:
        int    report;
        bool   lsq;
        double epsilon;
        
        float  minrgap;
        int    maxsplits;
        int    minchange;
        
        double delta;
        double adelta;
        double qdelta;
        double qdelta_req;
        float  minoverlap; //Fraction of area covered by line bb, so that char_box is included in line_box
        
        float  tr; //rejection threshold for the height of a box = tr*xheight
        int    offset; //Avg distance between words
        int    min_height;
        int    assign_boxes;
        
        bool   aggressive;

        typedef vecni<nparams> Parameters;
        double splitscale[nparams];
        Parameters all_params;
        Parameters empty_parameters;
        
        int do_shrink ;
        
        vec2i normalized(vec2i v) {
            interval a = atan2(v.y,v.x);
            return vec2i(cos(a),sin(a));
        }

        inline interval influence(bool lsq,interval d,double epsilon) {
            if(lsq) return sqinfluence(d,epsilon);
            else return rinfluence(d,epsilon);
        }
        
        typedef narray<int> Matches;
        narray<Rect> cboxes;
        narray<Rect> cboxes_all;
        narray<Rect> wboxes;
        narray<bool> used;
        narray<bool> used_all;
        int generation;
        int pagewidth, pageheight;
        
        bool final(interval q,const Parameters &p) {
            return p[0].width()<delta &&
                p[1].width()<adelta &&
                p[2].width()<delta;
        }
        
        struct TLState {
            short       generation;
            short       depth;
            short       rank;
            signed char splits;
            bool        splittable;
            interval    quality;
            float       priority;
            Parameters  params;
            Matches     matches;
            Rect        boundingBox;
            float       xheight;
            
            TLState();
            void set(CTextlineRAST &line,int depth,Parameters &params,
                     Matches &candidates,int splits);
            void reeval(CTextlineRAST &line);
            void update(CTextlineRAST &line, Matches &candidates);
            TextLine returnLine();
        };
	
        typedef counted<TLState> CState;
        heap<CState> queue;
        heap<CState> results_sorted; //Sort w.r.t y-intercept (params[0])
        narray<CState> results;
        autodel<CharStats> linestats;
        Matches all_matches;
        narray<Parameters> all_shrunk;
        
        struct line{
            float c,m,d; // c is y-intercept, m is slope, d is the line of descenders
            float start,end,top,bottom; // start and end of line segment
            float istart,iend; //actual start and end of line segment in the image
            float xheight;

            line() {}
            line(TextLine &tl);
            TextLine getTextLine();
        };
        
        void prepare();
        void make_substates(narray<CState> &substates,CState &state);
        int  wbox_intersection(CState &top);
        void search();
        int  wbox_intersection(line l, Rect wbox);    
        void extend_lines(narray<line> &lines, narray<Rect> &wboxes);    
        bool x_overlap (line a, line b);    
        bool separator_segment_found(line a, line b, narray<line> &lines);    
        void construct_graph(narray<line> &lines, narray<bool> &lines_dag);    
        int  id;
        narray<int> val;
        narray<int> ro_index;
    
        void visit (int k, narray<bool> &lines_dag);    
        void depth_first_search(narray<bool> &lines_dag);    
        void topological_sort(narray<bool> &lines_dag);    
    public:
        float  min_length;
        int    min_gap;
        double min_q;
        int    min_count;
        int    max_results;
        double max_slope;
        bool   use_whitespace;
        int    extend; 

        CTextlineRAST();
        void extract(narray<TextLine> &textlines, autodel<CharStats> &charstats);
        void extract(narray<TextLine> &textlines, narray<Rect> &columns,
                     autodel<CharStats>    &charstats);
        void rosort(narray<TextLine> &textlines, narray<Rect> &columns);
        void color(intarray &image,bytearray &in,narray<TextLine> &textlines);
    };
    CTextlineRAST *make_CTextlineRAST();

    ISegmentPage *make_SegmentPageByRAST();

}

#endif
