// -*- C++ -*-

// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-layout-rast - Extract textlines from a document image using RAST
// File: ocr-layout-rast.cc
// Purpose: perform layout analysis by RAST
// Responsible: Faisal Shafait (faisal.shafait@dfki.de)
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#include "iupr.h"
#include "io_png.h"
#include "io_pbm.h"
#include "img.h"
#include "autoinvert.h"
#include "ocr-layout-rast.h"

namespace iupr {

param_bool colorful("colorful",0,"output a colorful image map");
param_int  debug("debug",0,"print the intermediate results to stdout");

void invertbg(intarray &image){
    simple_recolor(image);
    for(int i=0; i<image.length1d(); i++)
        if(image.at1d(i) == 0x00800000)
            image.at1d(i) = 0x00ffffff;
    return;
}

struct SegmentPageByRAST : ISegmentPage {
    ~SegmentPageByRAST() {}

    const char *description() {
	return "segment characters by RAST\n";
    }

    void init(const char **argv) {
	// nothing to be done
    }
    
    void segment(intarray &image,bytearray &in_not_inverted) {
        const int zero   = 0;
        const int yellow = 0x00ffff00;
        bytearray in;
        copy(in, in_not_inverted);
        autoinvert(in);
        
        // Do connected component analysis
	intarray charimage;
	copy(charimage,in);
	label_components(charimage,false);

        // Clean non-text and noisy boxes and get character statistics
	narray<rectangle> bboxes;
	bounding_boxes(bboxes,charimage);
        if(bboxes.length()==0){
            makelike(image,in);
            fill(image,0x00ffffff);
            return ;
        }
        autodel<CharStats> charstats(make_CharStats());
        charstats->get_char_boxes(bboxes);
        charstats->calc_char_stats();
        if(debug>=2){
            charstats->print();
        }

        // Compute Whitespace Cover
        autodel<WhitespaceCover> whitespaces(make_WhitespaceCover(0,0,in.dim(0),in.dim(1)));
        narray<Rect> whitespaceboxes;
        whitespaces->compute(whitespaceboxes,charstats->char_boxes);

        // Find column separators
        autodel<ColSeparators> gutters(make_ColSeparators());
        narray<Rect> columns,colcandidates;
        gutters->find_gutters(colcandidates,whitespaceboxes,charstats);
        gutters->filter_overlaps(columns,colcandidates);
        if(debug){
            for(int i=0; i<columns.length();i++){
                printf("%d %d %d %d\n",columns[i].x0,columns[i].y0,
                       columns[i].x1,columns[i].y1);
            }
        }

        // Extract textlines
        autodel<CTextlineRAST> ctextline(make_CTextlineRAST());
        narray<TextLine> textlines;
        ctextline->min_q     = 2.0; // Minimum acceptable quality of a textline
        ctextline->min_count = 2;   // ---- number of characters in a textline
        ctextline->min_length= 30;  // ---- length in pixels of a textline
        ctextline->extract(textlines,columns,charstats);
        ctextline->rosort(textlines,columns);
        ctextline->color(image,in,textlines);
        if(debug){
            for(int i=0; i<textlines.length();i++){
                textlines[i].print();
            }
        }
        replace_values(image,zero,yellow);
    }
};

ISegmentPage *make_SegmentPageByRAST() {
    return new SegmentPageByRAST();
}


}

using namespace iupr;

#ifdef MAIN
int main(int argc, char **argv){
    try {
        USAGE((argc==3)||(argc==2),"Usage: ... input.png output.png\n");
        const char *outimage="out.png";
        if(argc==3){
            outimage=argv[2];
        }
	intarray image;
	read_png_rgb(image,stdio(argv[1],"r"),true);

	autodel<ISegmentPage> segmenter(make_SegmentPageByRAST());
	bytearray in;
	copy(in,image);
	segmenter->segment(image,in);

	if(colorful){ 
	    invertbg(image);
        }
	write_png_rgb(stdio(outimage,"w"),image);
    }
    catch(const char *oops) {
	fprintf(stderr,"oops: %s\n",oops);
    }
    
}
#endif
