// -*- C++ -*-

// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-layout-rast - Extract textlines from a document image using RAST
// File: rect.h
// Purpose: A structure for rectangle data type. Contains typical
// operations needed on rectangles.
// Responsible: Faisal Shafait (faisal.shafait@dfki.de)
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#ifndef H_RECT__
#define H_RECT__

#include "iupr.h"

using namespace iupr;

struct Rect {
    int x0,y0,x1,y1;
  /// Constructor
    Rect() {
	x0 = y0 = y1 = 0;
	x1 = -1;
    }
    Rect(int x0,int y0,int x1,int y1)
	:x0(x0),y0(y0),x1(x1),y1(y1) {}
    Rect(rectangle r)
	:x0(r.x0),y0(r.y0),x1(r.x1),y1(r.y1) {}
    int xcenter(){
        return (x0+x1)>>1;
    }
    int ycenter(){
        return (y0+y1)>>1;
    }
    int width() { return heaviside(x1-x0); }
    int height() { return heaviside(y1-y0); }
  /// Print rectangle coordinates to the terminal
    void print(FILE *stream=stdout) {
	fprintf(stream,"<Rect %d %d %d %d>\n",x0,y0,x1,y1);
    }
    int area() {
	return heaviside(x1-x0)*heaviside(y1-y0);
    }
    bool overlaps(const Rect &other) {
	return
	    x0<=other.x1 && x1>=other.x0 &&
	    y0<=other.y1 && y1>=other.y0;
    }
    Rect intersect(const Rect &other) {
	return Rect(max(x0,other.x0),max(y0,other.y0),min(x1,other.x1),min(y1,other.y1));
    }
    float fraction_covered_by(const Rect &other) {
	Rect isect = this->intersect(other);
        if(area())
	  return isect.area()/(float)area();
	else 
	  return -1;
    }
    float h_overlap(const Rect &other) {
	int ov0 = max(other.x0,x0);
	int ov1 = min(other.x1,x1);
        float ov  = (ov1 - ov0)/(float)width();

        return heaviside(ov);
    }
    float v_overlap(const Rect &other) {
	int ov0 = max(other.y0,y0);
	int ov1 = min(other.y1,y1);
        float ov  = (ov1 - ov0)/(float)height();

        return heaviside(ov);
    }
    float centricity(const Rect &other) {
	float width = x1-x0;
	float height = y1-y0;
	return
	    (other.x1-x0)/width *
	    (other.y1-y0)/height *
	    (x1-other.x0)/width *
	    (y1-other.y0)/height;
    }
  ///Expand current rectangle to include the point \f$(x,y)\f$.
    void include(int x,int y) {
	if(x1<x0) {
	    x0 = x1 = x;
	    y0 = y1 = y;
	} else {
	    x0 = min(x0,x);
	    y0 = min(y0,y);
	    x1 = max(x1,x);
	    y1 = max(y1,y);
	}
    }
    void include(const Rect &other) {
	include(other.x0,other.y0);
	include(other.x1,other.y1);
    }

    bool includes(int x,int y) {
	return (x >= x0 && x <= x1 && y >= y0 && y <= y1);
    }
  
    bool includes(float x,float y) {
	return (x >= x0 && x <= x1 && y >= y0 && y <= y1);
    }
  
    bool includes(const Rect &other) {
      return this->includes(other.x0,other.y0) && this->includes(other.x1,other.y1);
    }
  
  ///Expand the dimensions of current rectangle
    Rect dilated_by(int dx0,int dy0,int dx1,int dy1) {
	return Rect(x0-dx0,y0-dy0,x1+dx1,y1+dy1);
    }
  ///Return height / width
    float aspect() {
	return (y1-y0)/(float)(x1-x0);
    }
};

#endif
