// -*- C++ -*-

// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocr-layout-rast - Extract textlines from a document image using RAST
// File: struct.h
// Purpose: A collection of useful data structures
// Responsible: Faisal Shafait (faisal.shafait@dfki.de)
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de

#ifndef H_STRUCT__
#define H_STRUCT__


/////////////////////////////////////////////////////////////////////
///
/// \struct counted
/// Purpose: Reference counting class. Adds a variable refcount_ 
/// to the template class T, which counts the number of instances
/// of class T.
///  Also automatically allocates object as needed
///
//////////////////////////////////////////////////////////////////////

template <class T>
struct counted {
    struct TC : T {
	int refcount_;
    };
    TC *p;
    counted() {
	p = 0;
    }
    counted(const counted<T> &other) {
	other.incref();
	p = other.p;
    }
    counted(counted<T> &other) {
	other.incref();
	p = other.p;
    }
    ~counted() {
	decref();
	p = 0;
    }
    void operator=(counted<T> &other) {
	other.incref();
	decref();
	p = other.p;
    }
    void operator=(const counted<T> &other) {
	other.incref();
	decref();
	p = other.p;
    }
    void operator*=(counted<T> &other) {
	other.incref();
	decref();
	p = other.p;
	other.drop();
    }
    void operator*=(const counted<T> &other) {
	other.incref();
	decref();
	p = other.p;
	other.drop();
    }
    bool allocated() {
	return !p;
    }
    void allocate() {
	p = new TC();
	p->refcount_ = 1;
    }
    operator bool() {
	return !!p;
    }
    void drop() {
	decref();
	p = 0;
    }
    T &operator *() {
	if(!p) allocate();
	return *(T*)p;
    }
    T *operator->() {
	if(!p) allocate();
	return (T*)p;
    }
    operator T&() {
	if(!p) allocate();
	return *(T*)p;
    }
    operator T*() {
	if(!p) allocate();
	return (T*)p;
    }
    void incref() const {
	check();
	if(p) {
	    if(p->refcount_>10000000) abort();
	    if(p->refcount_<0) abort();
	    p->refcount_++;
	}
    }
    void decref() const {
	check();
	if(p) {
	    if(--p->refcount_==0) delete p;
	    ((counted<T>*)this)->p = 0;
	}
    }
    void check() const {
	if(!p) return;
	if(p->refcount_>10000000) abort();
	if(p->refcount_<0) abort();
    }
};


#endif
