// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz
// or its licensors, as applicable.
//
// You may not use this file except under the terms of the accompanying license.
//
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Project: ocr-adaptive
// File: tesseract.cc
// Purpose: interfaces to Tesseract
// Responsible: mezhirov
// Reviewer:
// Primary Repository:
// Web Sites: www.iupr.org, www.dfki.de

#ifndef UNSAFE
#define ISOLATE_TESSERACT
#endif

#include <signal.h>
#include <string>  // otherwise `baseapi.h' will break

// IUPR includes
#include "tesseract.h"
#include "ocr-utils.h"
#include "io_pbm.h"
#include "io_png.h"
#include "ocrinterfaces.h"
#include "objlist.h"
#include "coords.h"
#include "imglabels.h"
#include "imgops.h"
#include "line-info.h"
//#include "utils.h"

// Tess includes

#include "../tesseract/img.h" // that is to disambiguate it into Tesseract's header
#include "tordvars.h"
#include "control.h"
#include "tessvars.h"
#include "tessbox.h"
#include "tessedit.h"
#include "imgs.h"
#include "edgblob.h"
#include "makerow.h"
#include "wordseg.h"
#include "output.h"
#include "tstruct.h"
#include "tessout.h"
#include "tface.h"
#include "adaptmatch.h"
#include "baseapi.h"

// these Tess-defined macros interfere with IUPR names
#undef rectangle
#undef min
#undef max

enum {MIN_HEIGHT = 30};

FLOAT32 GetBestRatingFor(TBLOB *Blob, LINE_STATS *LineStats, CLASS_ID ClassId);

// We have to use char-level adaptation because otherwise
// someone should do forced alignment somewhere.
void AdaptToChar(TBLOB *Blob,
                 LINE_STATS *LineStats,
                 CLASS_ID ClassId,
                 FLOAT32 Threshold);

namespace iupr {


class TesseractWrapper;
TesseractWrapper *tesseract_singleton = NULL;


void oops_tesseract_died(int signal) {
    fprintf(stderr, "ERROR: got signal from Tesseract (bug in Tesseract?)\n");
    exit(1);
}

#ifdef ISOLATE_TESSERACT
static bool inside_tesseract = false;
static sighandler_t SIGSEGV_old_handler;
static sighandler_t SIGFPE_old_handler;
static sighandler_t SIGABRT_old_handler;
#endif

void enter_tesseract() {
#ifdef ISOLATE_TESSERACT
    ASSERT(!inside_tesseract);
    inside_tesseract = true;
    SIGSEGV_old_handler = signal(SIGSEGV, oops_tesseract_died);
    SIGFPE_old_handler  = signal(SIGFPE,  oops_tesseract_died);
    SIGABRT_old_handler = signal(SIGABRT, oops_tesseract_died);
#endif
}

void leave_tesseract() {
#ifdef ISOLATE_TESSERACT
    ASSERT(inside_tesseract);
    inside_tesseract = false;
    signal(SIGSEGV, SIGSEGV_old_handler);
    signal(SIGFPE,  SIGFPE_old_handler);
    signal(SIGABRT, SIGABRT_old_handler);
#endif
}


// ________________   passing bitmap and blocks to Tesseract   ________________

/// Load a bytearray into the Tesseract.
/// This is what Tesseract does when it loads a TIFF file.
template<class T> void tessy_set_image(narray<T> &image) {
    page_image.create(image.dim(0), image.dim(1), 1);
    IMAGELINE line;
    line.init(image.dim(0));
    for (int y = 0; y < image.dim(1); y++) {
        for (int x = 0; x < image.dim(0); x++) {
            line.pixels[x] = image(x,y) ? 1 : 0;
        }
        page_image.put_line(0, y, image.dim(0), &line, 0);
    }
}


// ______________________   passing a separate character   ____________________

ROW *tessy_make_ocrrow(float baseline, float xheight, float descender, float ascender) {
    INT32 xstarts[] = {-32000};
    double quad_coeffs[] = {0,0,baseline};
    return new ROW(
        1,
        xstarts,
        quad_coeffs,
        xheight,
        ascender - (baseline + xheight),
        descender - baseline,
        0,
        0
    );
}



/// Pass a character image to Tesseract and return a TBLOB *,
/// to be freed later with free_blob().
/// Side effect: page_image is rewritten.
TBLOB *tessy_make_blob(bytearray &image, float baseline, float xheight, float descender, float ascender) {
    tessy_set_image(image);
    BLOCK *block = new BLOCK ("a character",
                              TRUE, 
                              0, 0,
                              0, 0,
                              page_image.get_xsize(),
                              page_image.get_ysize());

    // Create C_BLOBs from the page
    extract_edges(NULL, &page_image, &page_image,
                  ICOORD(image.dim(0), image.dim(1)), block);

    // Create one PBLOB from all C_BLOBs
    C_BLOB_LIST *list = block->blob_list();
    C_BLOB_IT c_blob_it(list);
    PBLOB *pblob = new PBLOB; // will be (hopefully) deleted by the pblob_list
    for (c_blob_it.mark_cycle_pt();
        !c_blob_it.cycled_list();
         c_blob_it.forward()) {
        C_BLOB *c_blob = c_blob_it.data();
        PBLOB c_as_p(c_blob, baseline + xheight);
        merge_blobs(pblob, &c_as_p);
    }
    PBLOB_LIST *pblob_list = new PBLOB_LIST; // will be deleted by the word
    PBLOB_IT pblob_it(pblob_list);
    pblob_it.add_after_then_move(pblob);

    // Normalize PBLOB
    WERD word(pblob_list, 0, " ");
    ROW *row = tessy_make_ocrrow(baseline, xheight, descender, ascender);
    word.baseline_normalise(row);
    delete row;

    // Create a TBLOB from PBLOB
    return make_tess_blob(pblob, /* flatten: */ TRUE);
}

// _______________________   getting Tesseract output   _______________________


/// Convert Tess rectangle to IUPR one
/// These (-1)s are strange; but they work on c_blobs' bboxes.
rectangle tessy_rectangle(const BOX &b) {
    return rectangle(b.left() - 1, b.bottom() - 1, b.right() - 1, b.top() - 1);
}

/// \brief Get a bounding box of a PBLOB.
rectangle tessy_get_blob_bbox(PBLOB *blob) {
    OUTLINE_LIST *outlines = blob->out_list();
    OUTLINE_IT it(outlines);
    rectangle result;
    for(it.mark_cycle_pt(); !it.cycled_list(); it.forward()) {
        OUTLINE *outline = it.data();
        outline->compute_bb();
        result = result.inclusion(tessy_rectangle(outline->bounding_box()));
    }
    return result;
}

float rating_to_cost(float rating) {
    rating = 100 + rating;
    // cuddled that to save from coverage profiler 
    // (I have never seen ratings worse than -100, 
    //  but the check won't hurt)
    if (rating < 0) rating = 0;
    return rating;
}

/*
void tessy_prepare() {
    BOX box;
    TO_BLOCK_LIST land_blocks, port_blocks;

    tessy_mark_as_one_block();
    tessy_preprocess_blocks(box, land_blocks, port_blocks);
    tessy_mark_as_one_row(box, land_blocks, port_blocks);
}*/

// Almost a copy of make_tess_row() from ccmain/tstruct.cpp.
void tessy_fill_dummy_row(TEXTROW &tessrow, float baseline, float xheight, float descender, float ascender)
{
    tessrow.baseline.segments = 1;
    tessrow.baseline.xstarts[0] = -32767;
    tessrow.baseline.xstarts[1] = 32767;
    tessrow.baseline.quads[0].a = 0;
    tessrow.baseline.quads[0].b = 0;
    tessrow.baseline.quads[0].c = bln_baseline_offset;
    tessrow.xheight.segments = 1;
    tessrow.xheight.xstarts[0] = -32767;
    tessrow.xheight.xstarts[1] = 32767;
    tessrow.xheight.quads[0].a = 0;
    tessrow.xheight.quads[0].b = 0;
    tessrow.xheight.quads[0].c = bln_baseline_offset + bln_x_height;
    tessrow.lineheight = bln_x_height;
    tessrow.ascrise = bln_x_height * (ascender - (xheight + baseline)) / xheight;
    tessrow.descdrop = bln_x_height * (descender - baseline) / xheight;
}



void tessy_adapt(bytearray &image, int truth, float baseline, float xheight, float descender, float ascender) {
    LINE_STATS LineStats;
    TEXTROW row;
    tessy_fill_dummy_row(row, baseline, xheight, descender, ascender);
    GetLineStatsFromRow(&row, &LineStats); 
    
    TBLOB *blob = tessy_make_blob(image, baseline, xheight, descender, ascender);
    float threshold;
    int best_class = 0;
    float best_rating = -100;
    

    // Classify to get a raw choice.
    LIST result = AdaptiveClassifier(blob, NULL, &row);
    LIST p;
    for (p = result; p != NULL; p = p->next) {
        A_CHOICE *tesschoice = (A_CHOICE *) p->node;
        if (tesschoice->rating > best_rating) {
            best_rating = tesschoice->rating;
            best_class = tesschoice->string[0];
        }
    }

    if (truth == best_class)
        threshold = GoodAdaptiveMatch;
    else {
        /* the blob was incorrectly classified - find the rating threshold
           needed to create a template which will correct the error with
           some margin.  However, don't waste time trying to make
           templates which are too tight. */
        threshold = GetBestRatingFor(blob, &LineStats, truth);
        threshold *= .9;
        const float max_threshold = .125;
        const float min_threshold = .02;

        if (threshold > max_threshold)
            threshold = max_threshold;

        // I have cuddled the following line to set it out of the strike
        // of the coverage testing tool. I have no idea how to trigger
        // this situation nor I have any necessity to do it. --mezhirov
        if (threshold < min_threshold) threshold = min_threshold;
    }
    
    if (blob->outlines)
        AdaptToChar(blob, &LineStats, truth, threshold);
    free_blob(blob);
}


static int counter = 0;
inline double sqr(double x) {return x * x;}
class TesseractWrapper : 
    TessBaseAPI,
    public ISimpleRetrainableOCR /*, public IClassifyLineCharacter */ {
private:
    int pass;

    void pass_grayscale_image_to_tesseract(bytearray &image) {
        UINT8 *pixels = new UINT8[image.length1d()];
        for(int x=0;x<image.dim(0);x++) for(int y=0;y<image.dim(1);y++)
            pixels[(image.dim(1) - y - 1) * image.dim(0) + x] = image(x,y);

        CopyImageToTesseract(pixels, 1, image.dim(0), 0, 0, image.dim(0), image.dim(1));

        delete [] pixels;
    }

    void extract_result_from_PAGE_RES_IT(ucarray &str,
                                         narray<rectangle> &bboxes,
                                         floatarray &costs,
                                         PAGE_RES_IT &page_res_it) {
        str.clear();
        bboxes.clear();
        page_res_it.restart_page();
        
        int word_count = 0;
        while(page_res_it.word() != NULL) {
            WERD_RES *word = page_res_it.word();
            const char *s = word->best_choice->string().string();

            if (word_count) {
                str.push(' ');
                costs.push(0);
                bboxes.push();
            }
            rectangle bln_rect(0,0,-1,1);
            PBLOB_LIST *blobs = word->outword->blob_list();
            PBLOB_IT it(blobs);
            int len = strlen(s);
            int boxes_old_len = bboxes.length();
            for (int i = 0; i < len; i++) {
                PBLOB *blob = it.data();
                rectangle current = tessy_get_blob_bbox(blob);
                bln_rect = bln_rect.inclusion(current);
                bboxes.push(current);
                str.push(s[i]);
                costs.push(rating_to_cost(word->best_choice->rating()));
                it.forward();
            }

            // Find the word bbox before normalization.
            // Here we can't use the C_BLOB bboxes directly,
            // since connected letters are not yet cut.
            rectangle real_rect;
            C_BLOB_LIST *cblobs = word->word->cblob_list();
            C_BLOB_IT c_it(cblobs);
            for (c_it.mark_cycle_pt(); !c_it.cycled_list(); c_it.forward()) {
                C_BLOB *blob = c_it.data();
                //bboxes.push(tessy_rectangle(blob->bounding_box()));
                real_rect = real_rect.inclusion(tessy_rectangle(blob->bounding_box()));
            }

            // Denormalize boxes by transforming the bbox of the whole bln word
            // into the denorm bbox (`real_rect') of the whole word.
            double x_stretch = double(real_rect.width()) / bln_rect.width();
            double y_stretch = double(real_rect.height()) / bln_rect.height();
            for (int i = boxes_old_len; i < boxes_old_len + len; i++) {
                rectangle &r = bboxes[i];
                r.x0 = int(real_rect.x0 + x_stretch * (r.x0 - bln_rect.x0) + 0.5);
                r.x1 = int(real_rect.x0 + x_stretch * (r.x1 - bln_rect.x0) + 0.5);
                r.y0 = int(real_rect.y0 + y_stretch * (r.y0 - bln_rect.y0) + 0.5);
                r.y1 = int(real_rect.y0 + y_stretch * (r.y1 - bln_rect.y0) + 0.5);
            }
            
            page_res_it.forward();
            word_count++;
        }
    }


public:
    virtual const char *description() {
        return "a wrapper around Tesseract";
    }
    virtual void init(const char **argv=0) {
    }

    TesseractWrapper(const char *path_to_us) : pass(1) {
        if (!counter) {
            Init(path_to_us, NULL, NULL, false, 0, NULL);
            set_pass1();
        }
        counter++;
    }

    ~TesseractWrapper() {
        counter--;
        ClearAdaptiveClassifier();
        if (!counter) {
            End();
        }
    }

    virtual void start_recognizing() {
    }

    virtual void recognize_gray(ucarray &result,
                                floatarray &costs,
                                narray<rectangle> &bboxes,
                                bytearray &input_image) {
        
        enter_tesseract();
        bytearray image;
        copy(image, input_image);
        enum {PADDING = 3};
        pad_by(image, PADDING, PADDING, byte(255));
        pass_grayscale_image_to_tesseract(image);
        BLOCK_LIST blocks;
        FindLines(&blocks);

        
        narray<autodel<WERD> > bln_words;

        // Recognize all words
        PAGE_RES page_res(&blocks);
        PAGE_RES_IT page_res_it(&page_res);
        while (page_res_it.word () != NULL) {
            WERD_RES *word = page_res_it.word();
            ROW *row = page_res_it.row()->row;

            matcher_pass = 0;
            WERD *bln_word = make_bln_copy(word->word, row, row->x_height (),
                                           &word->denorm);
            bln_words.push() = bln_word;
            BLOB_CHOICE_LIST_CLIST blob_choices;

            if (pass == 1) {
                word->best_choice = tess_segment_pass1 (bln_word, &word->denorm,
                    tess_default_matcher,
                    word->raw_choice, &blob_choices,
                    word->outword);
            } else {
                word->best_choice = tess_segment_pass2 (bln_word, &word->denorm,
                    tess_default_matcher,
                    word->raw_choice, &blob_choices,
                    word->outword);
            }

            //classify_word_pass1 (page_res_it.word(), page_res_it.row()->row, 
            //FALSE, NULL, NULL);

            page_res_it.forward();
        }
        extract_result_from_PAGE_RES_IT(result, bboxes, costs, page_res_it);

        // Correct the padding.
        // Those +1/+2 corrections are necessary to un-Tesseract the boxes.
        for(int i = 0; i < bboxes.length(); i++) {
            bboxes[i].x0 += -PADDING+1;
            bboxes[i].y0 += -PADDING+2;
            bboxes[i].x1 += -PADDING+1;
            bboxes[i].y1 += -PADDING+2;
        }
        
        /*for(int i = 0; i < result.length(); i++) {
            printf("%c %d %d %d %d\n", result[i],
                    bboxes[i].x0, 
                    input_image.dim(1) - bboxes[i].y0,
                    bboxes[i].x1, 
                    input_image.dim(1) - bboxes[i].y1);
        }*/
        leave_tesseract();
    }
    

    virtual void recognize_binary(ucarray &result,floatarray &costs,narray<rectangle> &bboxes,bytearray &orig_image) {
        recognize_gray(result, costs, bboxes, orig_image);
    }

    virtual void start_training() {
        pass = 2;
    }
    
    virtual bool supports_char_training() {
        return true;
    }
    
    void train(ucarray &chars,intarray &orig_csegmentation) {
        intarray csegmentation;
        copy(csegmentation, orig_csegmentation);
        check_line_segmentation(csegmentation);
        white_to_black(csegmentation);
        check_line_segmentation(csegmentation,true);
        set_line_number(csegmentation, 0);
        int n = max(csegmentation);
        REQUIRE(chars.length() >= n);
        float intercept;
        float slope;
        float xheight;
        float ascender_rise;
        float descender_sink;
        bytearray line;
        
        if (!get_extended_line_info(intercept, slope, xheight,
                                    descender_sink, ascender_rise, csegmentation)) {
            return;
        }

        bytearray bitmap;
        for (int i = 0; i < n; i++) {
            set_pass1(); // because Tesseract adapts on pass 1
            makelike(bitmap, csegmentation);
            // int n = csegmentation.length1d();
            rectangle bbox(0,0,-1,-1);
            for (int x = 0; x < bitmap.dim(0); x++)
            for (int y = 0; y < bitmap.dim(1); y++) {
                if (csegmentation(x,y) == i + 1) {
                    bitmap(x,y) = 1;
                    bbox.include(x,y);
                } else {
                    bitmap(x,y) = 0;
                }
            }

            // Checking whether bbox is non-empty is a kluge.
            // It might be empty in the case of undersegmentation.
            // In this case, due to recoloring, only one 
            // character will actually receive the segment.
            // Note that this situation probably suggests
            // that we'd better not train on this word.
            // But we still do. FIXME?
            if (bbox.width() > 0 && bbox.height() > 0)
            {
                int center_x = (bbox.x0 + bbox.x1) / 2;
                float baseline = intercept + center_x * slope ;
                tessy_adapt(bitmap, chars[i], baseline, xheight,
                                baseline - descender_sink,
                                baseline + xheight + ascender_rise);
            }
        }
    }
    
    virtual void train_binary_chars(ucarray &chars,intarray &csegmentation) {
        enter_tesseract();
        train(chars, csegmentation);
        leave_tesseract();
    }

    void binarize_in_segmentation(intarray &segmentation, bytearray &gray_image) {
        CHECK(samedims(segmentation, gray_image));
        // should be passed a valid segmentation
        check_line_segmentation(segmentation,true);
        int n = segmentation.length1d();
        for(int i = 0; i < n; i++) {
            if(gray_image.at1d(i) < 128)
                segmentation.at1d(i) = 0xFFFFFF;
        }            
        // should return a valid segmentation
        check_line_segmentation(segmentation,true);
    }

    virtual void train_gray_chars(ucarray &chars,intarray &csegmentation,bytearray &image) {
        enter_tesseract();
        check_line_segmentation(csegmentation);
        intarray new_segmentation;
        copy(new_segmentation, csegmentation);
        binarize_in_segmentation(new_segmentation, image);
        check_line_segmentation(new_segmentation);
        train(chars, new_segmentation);
        leave_tesseract();
    }
    
    virtual bool supports_line_training() {
        return false;
    }
    virtual void train_binary(ucarray &chars,bytearray &bimage) {
        throw "TesseractWrapper: linewise training is not supported";
    }
    virtual void train_gray(ucarray &chars,bytearray &image) {
        throw "TesseractWrapper: linewise training is not supported";
    }

    /*virtual void score(
        intarray &classes,
        floatarray &scores,
        bytearray &image,
        int baseline,
        int xheight,
        int descender,
        int ascender) {
        //printf("tess> %d %d %d %d\n", baseline, xheight, descender, ascender);
        TBLOB *tess_blob = tessy_make_blob(image, baseline, xheight, descender, ascender);

        // Make a fake row
        TEXTROW tessrow;
        tessy_fill_dummy_row(tessrow, baseline, xheight, descender, ascender);

        LIST result = AdaptiveClassifier(tess_blob, NULL, &tessrow);
        free_blob(tess_blob);

        LIST p;
        for (p = result; p != NULL; p = p->next) {
            A_CHOICE *tesschoice = (A_CHOICE *) p->node;
            if (!tesschoice->string[0]) continue;
            classes.push(tesschoice->string[0]);
            //scores.push(exp(tesschoice->certainty * .1));
            //scores.push(exp(-tesschoice->rating * .1)); 
            double x = exp(-(tesschoice->rating + 5) * .03);

            // Now here's a KLUGE!
            double y = 1.3*exp(-80*sqr(x-0.43))+.6*exp(-90*sqr(x-0.65))+1.1*x;
            double p = x * y;
            if (p < 0) p = 0;
            if (p > 1) p = 1;
            scores.push(p); 
        }
    }*/

};


ISimpleRetrainableOCR *make_tesseract(const char *path_to_our_executable) {
    return new TesseractWrapper(path_to_our_executable);
}
//IClassifyLineCharacter *make_tesseract_classifier(const char *path_to_our_executable) {
//    return new TesseractWrapper(path_to_our_executable);
//}

}
