// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz
// or its licensors, as applicable.
//
// You may not use this file except under the terms of the accompanying license.
//
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Project:
// File: ocr-utils.h
// Purpose: miscelaneous routines; also adapters of ISimple... to ISegmented...
// Responsible: mezhirov
// Reviewer:
// Primary Repository:
// Web Sites: www.iupr.org, www.dfki.de


#ifndef h_utils_
#define h_utils_


// Defining directory separator
#if defined(WINDOWS) || defined (windows) || defined(_Windows)  \
 || defined(_WINDOWS) || defined(__WINDOWS__) || defined(WIN32) \
 || defined(_WIN32) || defined(__WIN32__) || defined(_WIN32_WCE)
#   ifndef WINDOWS
#       define WINDOWS
#   endif
#endif

#ifdef WINDOWS
#   define DIRSEP '\\'
#else
#   define DIRSEP '/'
#endif

#include "iupr.h"
#include "imgops.h"
#include "ocrinterfaces.h"
#include "coords.h"

namespace iupr {
    /// Simply 255 - array.
    void invert(bytearray &a);

    /// Copy the `src' to `dest', moving it by `shift_x' to the right and by `shift_y' up.
    /// The image must fit.
    void blit2d(bytearray &dest, const bytearray &src, int shift_x = 0, int shift_y = 0);

    /// Remove from the segmentation those pixels which are white in gray_image.
    void binarize_in_segmentation(intarray &segmentation, /* const */ bytearray &gray_image);

    float median(intarray &a);

    /// Estimate xheight given a slope and a segmentation.
    /// (That's an algorithm formerly used together with MLP).
    float estimate_xheight(intarray &seg, float slope);

    void plot_hist(FILE *stream, floatarray &hist);

    void get_line_info(float &baseline, float &xheight, float &descender, float &ascender, intarray &seg);

    /// Crop a 2D-array
    template<class T> void crop(narray<T> &result, narray<T> &source, int x, int y, int w, int h) {
        result.resize(w, h);
        for (int i = 0; i < w; i++) {
            for (int j = 0; j < h; j++)
                result(i, j) = source(x + i, y + j);
        }
    }
    
    /// Crop a 2D-array
    template<class T> void crop(narray<T> &result, narray<T> &source, rectangle r) {
        crop<T>(result, source, r.x0, r.y0, r.x1 - r.x0, r.y1 - r.y0);
    }

    template<class T> void crop(narray<T> &a, rectangle box) {
        narray<T> t;
        extract_subimage(t, a, box.x0, box.y0, box.x1, box.y1);
        move(a, t);
    }
    

    template<class T> void transpose(narray<T> &a) {
        narray<T> t;
        t.resize(a.dim(1), a.dim(0));
        for (int x = 0; x < a.dim(0); x++) {
            for (int y = 0; y < a.dim(1); y++)
                t(y,x) = a(x,y);
        }
        move(a, t);
    }

    template<class T> void replace_values(narray<T> &a, T from, T to) {
        for(int i = 0; i < a.length1d(); i++) {
            if (a.at1d(i) == from)
                a.at1d(i) = to;
        }
    }
    
    inline void white_to_black(intarray &a) {
        replace_values(a, 0xFFFFFF, 0);
    }
    
    inline void black_to_white(intarray &a) {
        replace_values(a, 0, 0xFFFFFF);
    }

    /// Set line number for all foreground pixels in a character segmentation.
    inline void set_line_number(intarray &a, int lnum) {
        lnum <<= 12;
        for(int i = 0; i < a.length1d(); i++) {
            if (a.at1d(i) && a.at1d(i) != 0xFFFFFF)
                a.at1d(i) = (a.at1d(i) & 0xFFF) | lnum;
        }
    }

    template<class T> void assert_all_items_nonempty(narray<narray<T> > &a) {
        for(int i=0;i<a.length();i++)
            ASSERT(a[i].length1d() > 0);
    }

    /// Unpack page segmentation into separate line masks with bounding boxes.
    void extract_lines(narray<bytearray> &lines,narray<rectangle> &rboxes,intarray &image);

    /// If the line is too small or too large, rescale it (with the mask)
    /// to a decent height (30-60 pixels).
    void rescale_if_needed(bytearray &bin_line, bytearray &gray_line);

    void debug_write(const char *s, int line, bytearray &a);
    void debug_write(const char *s, int line, intarray &a);
    
    const char *get_version_string();
    void set_version_string(const char *);
    
    /// Make a binary image from a line segmentation.
    void forget_segmentation(bytearray &image, intarray &segmentation);

    bool has_no_black_pixels(intarray &);

    void blit_segmentation_line(intarray &page,
                                rectangle bbox, 
                                intarray &line,
                                int line_no);

    // Enlarge segmentation and AND it with line_mask.
    // Don't pass binarized grayscale image as line_mask,
    // otherwise you might get debris not from the line.
    // (that means we cannot really call this from inside LineOCR)
    void normalize_segmentation(intarray &segmentation, bytearray &line_mask);

    void binarize_by_threshold(bytearray &image);
    int max_cnum(intarray &seg);

    // check whether cseg is a valid color segmentation of a line
    void check_line_segmentation(intarray &cseg,bool allow_zero=false,bool allow_gaps=false);

    // check whether pseg is a valid color segmentation of a page
    void check_page_segmentation(intarray &cseg,bool allow_zero=false);

    template<class T> bool contains_only(narray<T> &a, T value) {
        for(int i = 0; i < a.length1d(); i++) {
            if(a.at1d(i) != value)
                return false;
        }
        return true;
    }
    
    template<class T> bool contains_only(narray<T> &a, T value1, T value2) {
        for(int i = 0; i < a.length1d(); i++) {
            if(a.at1d(i) != value1 && a.at1d(i) != value2)
                return false;
        }
        return true;
    }
};

#endif
