// Copyright 2006 Deutsches Forschungszentrum fuer Kuenstliche Intelligenz 
// or its licensors, as applicable.
// 
// You may not use this file except under the terms of the accompanying license.
// 
// Licensed under the Apache License, Version 2.0 (the "License"); you
// may not use this file except in compliance with the License. You may
// obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// 
// Project: ocropus-cmd
// File: ocropus.cc
// Purpose: the main program file for OCRopus
// Responsible: mezhirov
// Reviewer: 
// Primary Repository: 
// Web Sites: www.iupr.org, www.dfki.de, www.ocropus.org

#include "iupr.h"
#include "imgio.h"
#include "coords.h"
#include "autoinvert.h"
#include "imglabels.h"
#include "triv-hocr.h"
#include "ocr-layout-rast.h"
#include "ocr-recognize-page.h"
#include "ocr-binarize-sauvola.h"
#include "langmod-ispell.h"
#include "langmod-aspell.h"
#include "langmod-shortest-path.h"
#include "segmentation.h"
#include "tesseract.h"
#include "ocr-utils.h"
#include "ocr-adapters.h"
#include "version.h"

#define TRAINING_THRESHOLD 30

using namespace iupr;

char version_string_buffer[100];
void init_version_string() {
/*    int rev;
    sscanf(version_string, "Revision: %d", &rev);
    sprintf(version_string_buffer, "Revision %d", rev);
    set_version_string(version_string_buffer);*/
    set_version_string(version_string);
}


void print_help() {
    puts("OCRopus - an open source OCR system");
    printf("%s\n", get_version_string());
    puts("Usage:");
    puts("    ocropus ocr <image> ...               - recognize given page(s)");
    puts("    ocropus ocr                           - same, reading file names from stdin");
    puts("    ocropus linerec <image>               - recognize a single line");
    puts("    ocropus linerecs <image>              - recognize many lines with adaptation");
    puts("    ocropus binarize <image> <result.png> - binarize");
    puts("    ocropus pageseg <image> <result.png>  - segment page into lines");
    puts("    ocropus lineseg <image> <result.png>  - oversegment a single line");
    puts("    ocropus extract-lines ...             - extract lines");
    puts("");
    puts("Options are passed through the environment only.");
    puts("Example: to get plain text output, use");
    puts("    hocr=0 ocropus ...");
    puts("Refer to the websites for more information:");
    puts("    http://www.ocropus.org");
    puts("    http://misc.iupr.org/docs/doku.php?id=pub:ocropus:start");
}



namespace iupr {
    param_bool spellcheck("spell",true,"applies spellchecking at the first pass");
    param_bool nocatch("nocatch",false,"don't catch exceptions");
    param_bool recolor("recolor",false,"recolor segmentation outputs (for easier display)");
    param_bool timing("timing",false,"output timing information");

    param_string binarize_name("binarize","sauvola","binarizer");
    param_string segment_page_name("segment_page","rast","page segmenter");
    param_string segment_line_name("segment_line","curved","line segmenter");
    param_string language_model_name("language_model","aspell","language_model");
    param_string line_ocr_name("line_ocr","tesseract","line ocr");
    param_string format("format",NULL,"input file format");
};


IBinarize *make_Binarize(const char *name) {
    if(!strcmp(name,"sauvola")) return make_BinarizeBySauvola();
    //if(!strcmp(name,"range")) return make_BinarizeByRange();
    throw "unknown binarizer";
}

ISegmentPage *make_SegmentPage(const char *name) {
    //if(!strcmp(name,"smear")) return make_SegmentPageBySmear();
    //if(!strcmp(name,"1cp")) return make_SegmentPageBy1CP();
    if(!strcmp(name,"rast")) return make_SegmentPageByRAST();
    throw "unknown page segmenter";
}

ISegmentLine *make_SegmentLine(const char *name) {
    //if(!strcmp(name,"proj")) return make_SegmentLineByProjection();
    //if(!strcmp(name,"ccs")) return make_SegmentLineByCCS();
    if(!strcmp(name,"curved")) return make_CurvedCutSegmenter();
    throw "unknown line segmenter";
}

ISegmentedRetrainableOCR *make_LineOCR(const char *name,char *path) {
    if(!strcmp(name,"tesseract")) {
        ISimpleRetrainableOCR *tesseract = make_tesseract(path);
        ISegmentedRetrainableOCR *pass2 = make_adapter(tesseract,TRAINING_THRESHOLD);
        return pass2;
    }
    throw "unknown line recognizer type";
}

void load_image_possibly_binarizing(bytearray &gray_line, bytearray &line_mask, const char *path) {
    read_image_gray(gray_line, path);
    // TODO: check if we have already binarized images around (?)
    copy(line_mask, gray_line);
    binarize_by_threshold(line_mask);
}

int cmd_extract_lines(intarray &pageseg,
                       const char *prefix,
                       int start_numbering = 1,
                       const char *gray_page = NULL,
                       const char *gray_prefix = NULL) {
    ASSERT(prefix);
    check_page_segmentation(pageseg);
    narray<bytearray> lines;
    narray<rectangle> bboxes; 
    extract_lines(lines,bboxes,pageseg);
    bytearray image;
    if(gray_page) {
        read_image_gray(image, gray_page);
    }
    assert_all_items_nonempty(lines);
    int prefix_len = strlen(prefix);
    for(int i = 0; i < lines.length(); i++) {
        char buf[prefix_len + 100];
        sprintf(buf, "%s%06d.png", prefix, i + start_numbering);
        write_png(stdio(buf, "wb"), lines[i]);
        if(gray_page) {
            sprintf(buf, "%s%06d.png", gray_prefix, i + start_numbering);
            bytearray line_image;
            extract_subimage(line_image, image,
                bboxes[i].x0, bboxes[i].y0, bboxes[i].x1, bboxes[i].y1);
            write_png(stdio(buf, "wb"), line_image);
        }
    }
    return lines.length() + start_numbering;
}

int real_main(int argc, char *argv[]) {
    char *executable = argv[0];
    argc--;
    argv++;
    if(!strcmp(argv[0],"copy")) {
        if(argc!=3) throw "usage ... copy input_image output.png";
        bytearray image;
        read_image_gray(image, argv[1], format);
        write_png(stdio(argv[2], "w"), image);
    } else if(!strcmp(argv[0],"binarize")) {
        if(argc!=3) throw "usage ... binarize input.png output.png";
        autodel<IBinarize> binarize(make_Binarize(binarize_name));
        bytearray image,out;
        floatarray temp;
        read_image_gray(image,argv[1],format);
        copy(temp,image);
        binarize->binarize(out,temp);
        write_png(stdio(argv[2],"w"),out);
    } else if(!strcmp(argv[0],"pageseg")) {
        if(argc!=3) throw "usage ... pageseg input.png output.png";
        autodel<ISegmentPage> segment(make_SegmentPage(segment_page_name));
        bytearray image;
        intarray out;
        read_image_gray(image,argv[1],format);
        binarize_by_threshold(image);
        segment->segment(out,image);
	check_page_segmentation(out);
        if(recolor) simple_recolor(out);
        write_png_rgb(stdio(argv[2],"w"),out);
    } else if(!strcmp(argv[0],"lineseg")) {
        if(argc!=3) throw "usage ... lineseg input.png output.png";
        autodel<ISegmentLine> segment(make_SegmentLine(segment_line_name));
        bytearray image;
        intarray out;
        read_image_gray(image,argv[1],format);
        segment->charseg(out,image);
        if(recolor) simple_recolor(out);
        write_png_rgb(stdio(argv[2],"w"),out);
    } else if(!strcmp(argv[0],"linerecs")) {
        if(argc==1) throw "usage ... linerecs <lines> ...";
        ucarray ocr_result;
        ucarray to_train;
        floatarray costs_to_train;
        intarray oversegmentation;
        intarray true_segmentation;
        bytearray line_mask;
        bytearray gray_line;

        autodel<ISegmentLine> segline(make_SegmentLine(segment_line_name));
        autodel<ISegmentedRetrainableOCR> lineOCR (make_LineOCR(line_ocr_name,executable));
        autodel<ILanguageModel> langmod1(make_AspellLanguageModel(make_ShortestPathLanguageModel()));
        autodel<ILanguageModel> langmod2(make_ShortestPathLanguageModel());

        // pass 1
        for(int i = 1; i < argc; i++) {
            load_image_possibly_binarizing(gray_line, line_mask, argv[i]);
            recognize_line(ocr_result, to_train, costs_to_train,
                    oversegmentation, true_segmentation,
                    line_mask, gray_line,
                    *segline, *langmod1, *lineOCR, i);
            lineOCR->train_gray(to_train, costs_to_train, true_segmentation, gray_line);
        }
        // pass 2
        for(int i = 1; i < argc; i++) {
            load_image_possibly_binarizing(gray_line, line_mask, argv[i]);
            recognize_line(ocr_result, to_train, costs_to_train,
                    oversegmentation, true_segmentation,
                    line_mask, gray_line,
                    *segline, *langmod2, *lineOCR, i);
            for(int k = 0; k < ocr_result.length(); k++)
                printf("%lc", ocr_result[k]);
            printf("\n");
        }
    } else if(!strcmp(argv[0],"linerec")) {
        if(argc!=2 && argc!=3) throw "usage ... linerec input.png [binary.png]";
        autodel<ISegmentLine> segment(make_SegmentLine(segment_line_name));
        autodel<ISegmentedRetrainableOCR> lineocr(make_LineOCR(line_ocr_name,executable));
        autodel<ILanguageModel> langmod(make_ShortestPathLanguageModel());
        bytearray image;
        bytearray gray_image;
        idmap ids;
        intarray segments;
        read_image_gray(gray_image,argv[1],format);
        if(argc == 3) {
            read_image_gray(image,argv[2],format);    
        } else {
            floatarray float_image;
            autodel<IBinarize> binarize(make_Binarize(binarize_name));
            copy(float_image, gray_image);
            binarize->binarize(image, float_image);
            autoinvert(image);
            invert(image);
        }
        narray<rectangle> bboxes;
        intarray tmp;
        copy(tmp, image);
        bounding_boxes(bboxes, tmp);
        rectangle r = bboxes[255];
        enum {M = 2};
        r.x0 -= M; r.y0 -= M;
        r.x1 += M; r.y1 += M;
        r.intersect(rectangle(0, 0, image.dim(0), image.dim(1)));
        crop(image, r);
        crop(gray_image, r);
        rescale_if_needed(image, gray_image);

        segment->charseg(segments,image);
        lineocr->recognize_gray(*langmod,ids,segments,gray_image);
        langmod->compute(1);
        if(langmod->nresults()<1) {
            fprintf(stderr,"no result\n");
            return 1;
        } else {
            narray<int> result;
            floatarray costs;
            intarray ids;
            intarray states;
            langmod->nbest(result,costs,ids,states,0);
            for(int i=0;i<result.length();i++)
                printf("%c",result[i]);
            printf("\n");
        }
    } else if(!strcmp(argv[0],"layout")) {
        if(argc < 3) throw "usage ... layout prefix <pages>...";
        autodel<IBinarize> binarize(make_Binarize(binarize_name));
        autodel<ISegmentPage> pageseg(make_SegmentPage(segment_page_name));
        int next_line_number = 1;
        for(int i = 2; i < argc; i++) {
            bytearray gray_image, bin_image;
            read_image_gray(gray_image,argv[i],format);

            floatarray temp;
            copy(temp,gray_image);
            binarize->binarize(bin_image, temp);
            temp.clear();

            intarray seg;
            pageseg->segment(seg, bin_image);
            
            next_line_number = cmd_extract_lines(seg, argv[1], next_line_number);
        }
    } else if(!strcmp(argv[0],"extract-lines")) {
        if(argc!=3 && argc!=5) throw "usage ... extract-lines input.png prefix [gray-page gray-prefix]";
        intarray pageseg;
        bytearray image;
        read_png_rgb(pageseg, stdio(argv[1], "rb"));
        cmd_extract_lines(pageseg, argv[2], 1, argc == 5 ? argv[3] : NULL,  argc == 5 ? argv[4] : NULL);
    } else if(!strcmp(argv[0],"ocr")) {
        autodel<IPageOCR> pageocr(make_PageOCR());

        pageocr->inject_binarizer(make_Binarize(binarize_name));
        pageocr->inject_page_segmenter(make_SegmentPage(segment_page_name));
        pageocr->inject_line_segmenter(make_SegmentLine(segment_line_name));
        pageocr->inject_pass2(make_LineOCR(line_ocr_name,executable));
        pageocr->inject_langmod1(make_AspellLanguageModel(make_ShortestPathLanguageModel()));
        pageocr->inject_langmod2(make_ShortestPathLanguageModel());
    
        narray<strbuf> files;
        if(argc>1) {
            for(int i=1;i<argc;i++) {
                //if (argv[i][0] != '-')
                files.push() = argv[i];
            }
        } else {
            char buf[10000];
            while(fgets(buf,sizeof buf,stdin)) {
                files.push() = buf;
            }
        }

        recognize_multiple_pages_to_stdout(files, *pageocr);

        if(timing) report_ocr_timings();
    } else if (!strcmp(argv[0],"help")) {
        print_help();
    } else {
        throw "unrecognized command (use something like \"ocropus ocr page.png\")";
    }
    return 0;
}

int main(int argc,char **argv) {
    init_version_string();

    if (argc == 1) {
        print_help();
        return 0;
    }

    if(nocatch) {
        return real_main(argc,argv);
    } else {
        try {
            return real_main(argc,argv);
        }
        catch(const char *s) {
            fprintf(stderr, "%s: %s\n", argv[0], s);
            exit(251);
        }
    }
}
